"""
Configuration of the alternatives system

Control the alternatives system

.. code-block:: jinja

  {% set my_hadoop_conf = '/opt/hadoop/conf' %}

  {{ my_hadoop_conf }}:
    file.directory

  hadoop-0.20-conf:
    alternatives.install:
      - name: hadoop-0.20-conf
      - link: /etc/hadoop-0.20/conf
      - path: {{ my_hadoop_conf }}
      - priority: 30
      - require:
        - file: {{ my_hadoop_conf }}

  hadoop-0.20-conf:
    alternatives.remove:
      - name: hadoop-0.20-conf
      - path: {{ my_hadoop_conf }}

"""

# Define a function alias in order not to shadow built-in's
__func_alias__ = {"set_": "set"}


def __virtual__():
    """
    Only load if alternatives execution module is available.
    """
    if "alternatives.auto" in __salt__:
        return True
    return (False, "alernatives module could not be loaded")


def install(name, link, path, priority):
    """
    Install new alternative for defined <name>

    name
        is the master name for this link group
        (e.g. pager)

    link
        is the symlink pointing to /etc/alternatives/<name>.
        (e.g. /usr/bin/pager)

    path
        is the location of the new alternative target.
        NB: This file / directory must already exist.
        (e.g. /usr/bin/less)

    priority
        is an integer; options with higher numbers have higher priority in
        automatic mode.
    """
    ret = {
        "name": name,
        "link": link,
        "path": path,
        "priority": priority,
        "result": True,
        "changes": {},
        "comment": "",
    }

    if __salt__["alternatives.check_exists"](name, path):
        ret["comment"] = "Alternative {} for {} is already registered".format(
            path, name
        )
    else:
        if __opts__["test"]:
            ret["comment"] = (
                "Alternative will be set for {} to {} with priority {}".format(
                    name, path, priority
                )
            )
            ret["result"] = None
            return ret

        out = __salt__["alternatives.install"](name, link, path, priority)
        if __salt__["alternatives.check_exists"](name, path):
            if __salt__["alternatives.check_installed"](name, path):
                ret["comment"] = (
                    "Alternative for {} set to path {} with priority {}".format(
                        name, path, priority
                    )
                )
            else:
                ret["comment"] = (
                    "Alternative {} for {} registered with priority {} and "
                    "not set to default".format(path, name, priority)
                )
            ret["changes"] = {
                "name": name,
                "link": link,
                "path": path,
                "priority": priority,
            }
        else:
            ret["result"] = False
            ret["comment"] = f"Alternative for {name} not installed: {out}"

    return ret


def remove(name, path):
    """
    Removes installed alternative for defined <name> and <path>
    or fallback to default alternative, if some defined before.

    name
        is the master name for this link group
        (e.g. pager)

    path
        is the location of one of the alternative target files.
        (e.g. /usr/bin/less)
    """
    ret = {"name": name, "path": path, "result": True, "changes": {}, "comment": ""}

    isinstalled = __salt__["alternatives.check_exists"](name, path)
    if isinstalled:
        if __opts__["test"]:
            ret["comment"] = f"Alternative for {name} will be removed"
            ret["result"] = None
            return ret
        __salt__["alternatives.remove"](name, path)
        current = __salt__["alternatives.show_current"](name)
        if current:
            ret["result"] = True
            ret["comment"] = (
                "Alternative for {} removed. Falling back to path {}".format(
                    name, current
                )
            )
            ret["changes"] = {"path": current}
            return ret

        ret["comment"] = f"Alternative for {name} removed"
        ret["changes"] = {}
        return ret

    current = __salt__["alternatives.show_current"](name)
    if current:
        ret["result"] = True
        ret["comment"] = "Alternative for {} is set to it's default path {}".format(
            name, current
        )
        return ret

    ret["result"] = False
    ret["comment"] = f"Alternative for {name} doesn't exist"

    return ret


def auto(name):
    """
    .. versionadded:: 0.17.0

    Instruct alternatives to use the highest priority
    path for <name>

    name
        is the master name for this link group
        (e.g. pager)

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    display = __salt__["alternatives.display"](name)
    line = display.splitlines()[0]
    if line.endswith(" auto mode"):
        ret["comment"] = f"{name} already in auto mode"
        return ret

    if __opts__["test"]:
        ret["comment"] = f"{name} will be put in auto mode"
        ret["result"] = None
        return ret
    ret["changes"]["result"] = __salt__["alternatives.auto"](name)
    return ret


def set_(name, path):
    """
    .. versionadded:: 0.17.0

    Sets alternative for <name> to <path>, if <path> is defined
    as an alternative for <name>.

    name
        is the master name for this link group
        (e.g. pager)

    path
        is the location of one of the alternative target files.
        (e.g. /usr/bin/less)

    .. code-block:: yaml

        foo:
          alternatives.set:
            - path: /usr/bin/foo-2.0
    """
    ret = {"name": name, "path": path, "result": True, "changes": {}, "comment": ""}

    current = __salt__["alternatives.show_current"](name)
    if current == path:
        ret["comment"] = f"Alternative for {name} already set to {path}"
        return ret

    display = __salt__["alternatives.display"](name)
    isinstalled = False
    for line in display.splitlines():
        if line.startswith(path):
            isinstalled = True
            break

    if isinstalled:
        if __opts__["test"]:
            ret["comment"] = "Alternative for {} will be set to path {}".format(
                name, path
            )
            ret["result"] = None
            return ret
        __salt__["alternatives.set"](name, path)
        current = __salt__["alternatives.show_current"](name)
        if current == path:
            ret["comment"] = f"Alternative for {name} set to path {current}"
            ret["changes"] = {"path": current}
        else:
            ret["comment"] = f"Alternative for {name} not updated"

        return ret

    else:
        ret["result"] = False
        ret["comment"] = f"Alternative {path} for {name} doesn't exist"

    return ret



"""
    salt.serializers.yamlex
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    YAMLEX is a format that allows for things like sls files to be
    more intuitive.

    It's an extension of YAML that implements all the salt magic:
    - it implies omap for any dict like.
    - it implies that string like data are str, not unicode
    - ...

    For example, the file `states.sls` has this contents:

    .. code-block:: yaml

        foo:
          bar: 42
          baz: [1, 2, 3]

    The file can be parsed into Python like this

    .. code-block:: python

        from salt.serializers import yamlex

        with open('state.sls', 'r') as stream:
            obj = yamlex.deserialize(stream)

    Check that ``obj`` is an OrderedDict

    .. code-block:: python

        from collections import OrderedDict

        assert isinstance(obj, dict)
        assert isinstance(obj, OrderedDict)


    yamlex `__repr__` and `__str__` objects' methods render YAML understandable
    string. It means that they are template friendly.


    .. code-block:: python

        print '{0}'.format(obj)

    returns:

    ::

        {foo: {bar: 42, baz: [1, 2, 3]}}

    and they are still valid YAML:

    .. code-block:: python

        from salt.serializers import yaml
        yml_obj = yaml.deserialize(str(obj))
        assert yml_obj == obj

    yamlex implements also custom tags:

    !aggregate

         this tag allows structures aggregation.

        For example:


        .. code-block:: yaml

            placeholder: !aggregate foo
            placeholder: !aggregate bar
            placeholder: !aggregate baz

        is rendered as

        .. code-block:: yaml

            placeholder: [foo, bar, baz]

    !reset

         this tag flushes the computing value.

        .. code-block:: yaml

            placeholder: {!aggregate foo: {foo: 42}}
            placeholder: {!aggregate foo: {bar: null}}
            !reset placeholder: {!aggregate foo: {baz: inga}}

        is roughly equivalent to

        .. code-block:: yaml

            placeholder: {!aggregate foo: {baz: inga}}

    Document is defacto an aggregate mapping.
"""

# pylint: disable=invalid-name,no-member,missing-docstring,no-self-use
# pylint: disable=too-few-public-methods,too-many-public-methods


import copy
import datetime
import logging
from collections import OrderedDict

import yaml
from yaml.constructor import ConstructorError
from yaml.nodes import MappingNode
from yaml.scanner import ScannerError

from salt.serializers import DeserializationError, SerializationError
from salt.utils.aggregation import Map, Sequence, aggregate

__all__ = ["deserialize", "serialize", "available"]

log = logging.getLogger(__name__)

available = True

# prefer C bindings over python when available
BaseLoader = getattr(yaml, "CSafeLoader", yaml.SafeLoader)
# CSafeDumper causes repr errors in python3, so use the pure Python one
try:
    # Depending on how PyYAML was built, yaml.SafeDumper may actually be
    # yaml.dumper.SafeDumper.
    BaseDumper = yaml.dumper.SafeDumper
except AttributeError:
    # Here just in case, but yaml.dumper.SafeDumper should always exist
    BaseDumper = yaml.SafeDumper

ERROR_MAP = {
    "found character '\\t' that cannot start any token": "Illegal tab character"
}


def deserialize(stream_or_string, **options):
    """
    Deserialize any string of stream like object into a Python data structure.

    :param stream_or_string: stream or string to deserialize.
    :param options: options given to lower yaml module.
    """

    options.setdefault("Loader", Loader)
    try:
        return yaml.load(stream_or_string, **options)
    except ScannerError as error:
        log.exception("Error encountered while deserializing")
        err_type = ERROR_MAP.get(error.problem, "Unknown yaml render error")
        line_num = error.problem_mark.line + 1
        raise DeserializationError(err_type, line_num, error.problem_mark.buffer)
    except ConstructorError as error:
        log.exception("Error encountered while deserializing")
        raise DeserializationError(error)
    except Exception as error:  # pylint: disable=broad-except
        log.exception("Error encountered while deserializing")
        raise DeserializationError(error)


def serialize(obj, **options):
    """
    Serialize Python data to YAML.

    :param obj: the data structure to serialize
    :param options: options given to lower yaml module.
    """

    options.setdefault("Dumper", Dumper)
    options.setdefault("default_flow_style", None)
    try:
        response = yaml.dump(obj, **options)
        if response.endswith("\n...\n"):
            return response[:-5]
        if response.endswith("\n"):
            return response[:-1]
        return response
    except Exception as error:  # pylint: disable=broad-except
        log.exception("Error encountered while serializing")
        raise SerializationError(error)


class Loader(BaseLoader):  # pylint: disable=W0232
    """
    Create a custom YAML loader that uses the custom constructor. This allows
    for the YAML loading defaults to be manipulated based on needs within salt
    to make things like sls file more intuitive.
    """

    DEFAULT_SCALAR_TAG = "tag:yaml.org,2002:str"
    DEFAULT_SEQUENCE_TAG = "tag:yaml.org,2002:seq"
    DEFAULT_MAPPING_TAG = "tag:yaml.org,2002:omap"

    def compose_document(self):
        node = BaseLoader.compose_document(self)
        node.tag = "!aggregate"
        return node

    def construct_yaml_omap(self, node):
        """
        Build the SLSMap
        """
        sls_map = SLSMap()
        if not isinstance(node, MappingNode):
            raise ConstructorError(
                None,
                None,
                f"expected a mapping node, but found {node.id}",
                node.start_mark,
            )

        self.flatten_mapping(node)

        for key_node, value_node in node.value:

            # !reset instruction applies on document only.
            # It tells to reset previous decoded value for this present key.
            reset = key_node.tag == "!reset"

            # even if !aggregate tag apply only to values and not keys
            # it's a reason to act as a such nazi.
            if key_node.tag == "!aggregate":
                log.warning("!aggregate applies on values only, not on keys")
                value_node.tag = key_node.tag
                key_node.tag = self.resolve_sls_tag(key_node)[0]

            key = self.construct_object(key_node, deep=False)
            try:
                hash(key)
            except TypeError:
                raise ConstructorError(
                    "While constructing a mapping {} found unacceptable key {}".format(
                        node.start_mark, key_node.start_mark
                    )
                )
            value = self.construct_object(value_node, deep=False)
            if key in sls_map and not reset:
                value = merge_recursive(sls_map[key], value)
            sls_map[key] = value
        return sls_map

    def construct_sls_str(self, node):
        """
        Build the SLSString.
        """

        # Ensure obj is str, not py2 unicode or py3 bytes
        obj = self.construct_scalar(node)
        return SLSString(obj)

    def construct_sls_int(self, node):
        """
        Verify integers and pass them in correctly is they are declared
        as octal
        """
        if node.value == "0":
            pass
        elif node.value.startswith("0") and not node.value.startswith(("0b", "0x")):
            node.value = node.value.lstrip("0")
            # If value was all zeros, node.value would have been reduced to
            # an empty string. Change it to '0'.
            if node.value == "":
                node.value = "0"
        return int(node.value)

    def construct_sls_aggregate(self, node):
        try:
            tag, deep = self.resolve_sls_tag(node)
        except Exception:  # pylint: disable=broad-except
            raise ConstructorError("unable to build reset")

        node = copy.copy(node)
        node.tag = tag
        obj = self.construct_object(node, deep)
        if obj is None:
            return AggregatedSequence()
        elif tag == self.DEFAULT_MAPPING_TAG:
            return AggregatedMap(obj)
        elif tag == self.DEFAULT_SEQUENCE_TAG:
            return AggregatedSequence(obj)
        return AggregatedSequence([obj])

    def construct_sls_reset(self, node):
        try:
            tag, deep = self.resolve_sls_tag(node)
        except Exception:  # pylint: disable=broad-except
            raise ConstructorError("unable to build reset")

        node = copy.copy(node)
        node.tag = tag

        return self.construct_object(node, deep)

    def resolve_sls_tag(self, node):
        if isinstance(node, yaml.nodes.ScalarNode):
            # search implicit tag
            tag = self.resolve(yaml.nodes.ScalarNode, node.value, [True, True])
            deep = False
        elif isinstance(node, yaml.nodes.SequenceNode):
            tag = self.DEFAULT_SEQUENCE_TAG
            deep = True
        elif isinstance(node, yaml.nodes.MappingNode):
            tag = self.DEFAULT_MAPPING_TAG
            deep = True
        else:
            raise ConstructorError("unable to resolve tag")
        return tag, deep


Loader.add_constructor("!aggregate", Loader.construct_sls_aggregate)  # custom type
Loader.add_constructor("!reset", Loader.construct_sls_reset)  # custom type
Loader.add_constructor(
    "tag:yaml.org,2002:omap", Loader.construct_yaml_omap
)  # our overwrite
Loader.add_constructor(
    "tag:yaml.org,2002:str", Loader.construct_sls_str
)  # our overwrite
Loader.add_constructor(
    "tag:yaml.org,2002:int", Loader.construct_sls_int
)  # our overwrite
Loader.add_multi_constructor("tag:yaml.org,2002:null", Loader.construct_yaml_null)
Loader.add_multi_constructor("tag:yaml.org,2002:bool", Loader.construct_yaml_bool)
Loader.add_multi_constructor("tag:yaml.org,2002:float", Loader.construct_yaml_float)
Loader.add_multi_constructor("tag:yaml.org,2002:binary", Loader.construct_yaml_binary)
Loader.add_multi_constructor(
    "tag:yaml.org,2002:timestamp", Loader.construct_yaml_timestamp
)
Loader.add_multi_constructor("tag:yaml.org,2002:pairs", Loader.construct_yaml_pairs)
Loader.add_multi_constructor("tag:yaml.org,2002:set", Loader.construct_yaml_set)
Loader.add_multi_constructor("tag:yaml.org,2002:seq", Loader.construct_yaml_seq)
Loader.add_multi_constructor("tag:yaml.org,2002:map", Loader.construct_yaml_map)


class SLSMap(OrderedDict):
    """
    Ensures that dict str() and repr() are YAML friendly.

    .. code-block:: python

        >>> mapping = OrderedDict([('a', 'b'), ('c', None)])
        >>> print mapping
        OrderedDict([('a', 'b'), ('c', None)])

        >>> sls_map = SLSMap(mapping)
        >>> print sls_map.__str__()
        {a: b, c: null}

    """

    def __str__(self):
        return serialize(self, default_flow_style=True)

    def __repr__(self, _repr_running=None):
        return serialize(self, default_flow_style=True)


class SLSString(str):
    """
    Ensures that str str() and repr() are YAML friendly.

    .. code-block:: python

        >>> scalar = str('foo')
        >>> print 'foo'
        foo

        >>> sls_scalar = SLSString(scalar)
        >>> print sls_scalar
        "foo"

    """

    def __str__(self):
        return serialize(self, default_style='"')

    def __repr__(self):
        return serialize(self, default_style='"')


class AggregatedMap(SLSMap, Map):
    pass


class AggregatedSequence(Sequence):
    pass


class Dumper(BaseDumper):  # pylint: disable=W0232
    """
    sls dumper.
    """

    def represent_odict(self, data):
        return self.represent_mapping("tag:yaml.org,2002:map", list(data.items()))


Dumper.add_multi_representer(type(None), Dumper.represent_none)
Dumper.add_multi_representer(bytes, Dumper.represent_binary)
Dumper.add_multi_representer(str, Dumper.represent_str)
Dumper.add_multi_representer(bool, Dumper.represent_bool)
Dumper.add_multi_representer(int, Dumper.represent_int)
Dumper.add_multi_representer(float, Dumper.represent_float)
Dumper.add_multi_representer(list, Dumper.represent_list)
Dumper.add_multi_representer(tuple, Dumper.represent_list)
Dumper.add_multi_representer(
    dict, Dumper.represent_odict
)  # make every dict like obj to be represented as a map
Dumper.add_multi_representer(set, Dumper.represent_set)
Dumper.add_multi_representer(datetime.date, Dumper.represent_date)
Dumper.add_multi_representer(datetime.datetime, Dumper.represent_datetime)
Dumper.add_multi_representer(None, Dumper.represent_undefined)


def merge_recursive(obj_a, obj_b, level=False):
    """
    Merge obj_b into obj_a.
    """
    return aggregate(
        obj_a, obj_b, level, map_class=AggregatedMap, sequence_class=AggregatedSequence
    )



"""
Tests for logrotate config
"""

import pathlib

import packaging.version
import pytest

pytestmark = [
    pytest.mark.skip_unless_on_linux,
]


@pytest.fixture
def logrotate_config_file(grains):
    """
    Fixture for logrotate config file path
    """
    if grains["os_family"] == "RedHat":
        return pathlib.Path("/etc/logrotate.d", "salt")
    elif grains["os_family"] == "Debian":
        return pathlib.Path("/etc/logrotate.d", "salt-common")


def test_logrotate_config(logrotate_config_file):
    """
    Test that logrotate config has been installed in correctly
    """
    assert logrotate_config_file.is_file()
    assert logrotate_config_file.owner() == "root"
    assert logrotate_config_file.group() == "root"


def test_issue_65231_etc_logrotate_salt_dir_removed(install_salt):
    """
    Test that /etc/logrotate.d/salt is not a directory
    """
    if install_salt.prev_version and packaging.version.parse(
        install_salt.prev_version
    ) <= packaging.version.parse("3006.4"):
        pytest.skip("Testing a downgrade to 3006.4, do not run")

    path = pathlib.Path("/etc/logrotate.d/salt")
    if path.exists():
        assert path.is_dir() is False



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>

    Test cases for salt.modules.event
"""

import pytest

import salt.modules.event as event
import salt.utils.event
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {event: {"__opts__": minion_opts}}


def test_fire_master():
    """
    Test for Fire an event off up to the master server
    """
    with patch("salt.crypt.SAuth") as salt_crypt_sauth, patch(
        "salt.channel.client.ReqChannel.factory"
    ) as salt_transport_channel_factory:

        preload = {
            "id": "id",
            "tag": "tag",
            "data": "data",
            "tok": "salt",
            "cmd": "_minion_event",
        }

        with patch.dict(
            event.__opts__,
            {"transport": "A", "master_uri": "localhost", "local": False},
        ):
            with patch.object(salt_crypt_sauth, "gen_token", return_value="tok"):
                with patch.object(
                    salt_transport_channel_factory, "send", return_value=None
                ):
                    assert event.fire_master("data", "tag", preload)

        with patch.dict(event.__opts__, {"transport": "A", "local": False}):
            with patch.object(
                salt.utils.event.MinionEvent,
                "fire_event",
                side_effect=Exception("foo"),
            ):
                assert not event.fire_master("data", "tag")


def test_fire():
    """
    Test to fire an event on the local minion event bus.
    Data must be formed as a dict.
    """
    with patch("salt.utils.event") as salt_utils_event:
        with patch.object(salt_utils_event, "get_event") as mock:
            mock.fire_event = MagicMock(return_value=True)
            assert event.fire("data", "tag")


def test_send():
    """
    Test for Send an event to the Salt Master
    """
    with patch.object(event, "fire_master", return_value="B"):
        assert event.send("tag") == "B"


def test_send_use_master_when_local_false():
    """
    Test for Send an event when opts has use_master_when_local and its False
    """
    patch_master_opts = patch.dict(event.__opts__, {"use_master_when_local": False})
    patch_file_client = patch.dict(event.__opts__, {"file_client": "local"})
    with patch.object(event, "fire", return_value="B") as patch_send:
        with patch_master_opts, patch_file_client, patch_send:
            assert event.send("tag") == "B"
            patch_send.assert_called_once()


def test_send_use_master_when_local_true():
    """
    Test for Send an event when opts has use_master_when_local and its True
    """
    patch_master_opts = patch.dict(event.__opts__, {"use_master_when_local": True})
    patch_file_client = patch.dict(event.__opts__, {"file_client": "local"})
    with patch.object(event, "fire_master", return_value="B") as patch_send:
        with patch_master_opts, patch_file_client, patch_send:
            assert event.send("tag") == "B"
            patch_send.assert_called_once()



"""
Tests for salt.modules.linux_sysctl module

:codeauthor: jmoney <justin@saltstack.com>
"""

import os

import pytest

import salt.modules.linux_sysctl as linux_sysctl
import salt.modules.systemd_service as systemd
from salt.exceptions import CommandExecutionError
from salt.utils.files import fopen
from tests.support.mock import MagicMock, mock_open, patch

pytestmark = [
    pytest.mark.skip_on_windows(reason="sysctl not available on Windows"),
]


@pytest.fixture
def configure_loader_modules():
    return {linux_sysctl: {}, systemd: {}}


def test_get():
    """
    Tests the return of get function
    """
    mock_cmd = MagicMock(return_value=1)
    which_mock = MagicMock(return_value="/usr/sbin/sysctl")
    with patch("salt.utils.path.which", which_mock):
        with patch.dict(linux_sysctl.__salt__, {"cmd.run": mock_cmd}):
            assert linux_sysctl.get("net.ipv4.ip_forward") == 1
    mock_cmd.assert_called_once_with(
        ["/usr/sbin/sysctl", "-n", "net.ipv4.ip_forward"], python_shell=False
    )


def test_show():
    """
    Tests the return of show function
    """
    mock_cmd = MagicMock(
        return_value="""\
kernel.core_pattern = |/usr/share/kdump-tools/dump-core %p %s %t %e
kernel.printk = 3 4 1 3
net.ipv4.ip_forward = 1
net.ipv4.tcp_rmem = 4096	131072	6291456
"""
    )
    which_mock = MagicMock(return_value="/usr/sbin/sysctl")
    with patch("salt.utils.path.which", which_mock):
        with patch.dict(linux_sysctl.__salt__, {"cmd.run_stdout": mock_cmd}):
            assert linux_sysctl.show() == {
                "kernel.core_pattern": "|/usr/share/kdump-tools/dump-core %p %s %t %e",
                "kernel.printk": "3 4 1 3",
                "net.ipv4.ip_forward": "1",
                "net.ipv4.tcp_rmem": "4096\t131072\t6291456",
            }
    mock_cmd.assert_called_once_with(
        ["/usr/sbin/sysctl", "-a"], output_loglevel="trace"
    )


def test_show_config_file(tmp_path):
    """
    Tests the return of show function for a given file
    """
    config = str(tmp_path / "sysctl.conf")
    with fopen(config, "w", encoding="utf-8") as config_file:
        config_file.write(
            """\
# Use dump-core from kdump-tools Debian package.
kernel.core_pattern = |/usr/share/kdump-tools/dump-core %p %s %t %e
 # Stop low-level messages on console = less logging
 kernel.printk  = 3 4 1 3

net.ipv4.ip_forward=1
net.ipv4.tcp_rmem	=	4096	131072	6291456
"""
        )
    assert linux_sysctl.show(config) == {
        "kernel.core_pattern": "|/usr/share/kdump-tools/dump-core %p %s %t %e",
        "kernel.printk": "3 4 1 3",
        "net.ipv4.ip_forward": "1",
        "net.ipv4.tcp_rmem": "4096\t131072\t6291456",
    }


def test_get_no_sysctl_binary():
    """
    Tests the failure of get function when no binary exists
    """
    with patch("salt.utils.path.which", MagicMock(return_value=None)):
        with pytest.raises(CommandExecutionError):
            linux_sysctl.get("net.ipv4.ip_forward")


def test_assign_proc_sys_failed():
    """
    Tests if /proc/sys/<kernel-subsystem> exists or not
    """
    with patch("os.path.exists", MagicMock(return_value=False)):
        cmd = {
            "pid": 1337,
            "retcode": 0,
            "stderr": "",
            "stdout": "net.ipv4.ip_forward = 1",
        }
        mock_cmd = MagicMock(return_value=cmd)
        with patch.dict(linux_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
            with pytest.raises(CommandExecutionError):
                linux_sysctl.assign("net.ipv4.ip_forward", 1)
        mock_cmd.assert_not_called()


def test_assign_cmd_failed():
    """
    Tests if the assignment was successful or not
    """
    with patch("os.path.exists", MagicMock(return_value=True)):
        cmd = {
            "pid": 1337,
            "retcode": 0,
            "stderr": 'sysctl: setting key "net.ipv4.ip_forward": Invalid argument',
            "stdout": "net.ipv4.ip_forward = backward",
        }
        mock_cmd = MagicMock(return_value=cmd)
        which_mock = MagicMock(return_value="/usr/sbin/sysctl")
        with patch("salt.utils.path.which", which_mock):
            with patch.dict(linux_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
                with pytest.raises(CommandExecutionError):
                    linux_sysctl.assign("net.ipv4.ip_forward", "backward")
        mock_cmd.assert_called_once_with(
            ["/usr/sbin/sysctl", "-w", "net.ipv4.ip_forward=backward"],
            python_shell=False,
        )


def test_assign_success():
    """
    Tests the return of successful assign function
    """
    with patch("os.path.exists", MagicMock(return_value=True)):
        cmd = {
            "pid": 1337,
            "retcode": 0,
            "stderr": "",
            "stdout": "net.ipv4.ip_forward = 1",
        }
        ret = {"net.ipv4.ip_forward": "1"}
        mock_cmd = MagicMock(return_value=cmd)
        which_mock = MagicMock(return_value="/usr/sbin/sysctl")
        with patch("salt.utils.path.which", which_mock):
            with patch.dict(linux_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
                assert linux_sysctl.assign("net.ipv4.ip_forward", 1) == ret
        mock_cmd.assert_called_once_with(
            ["/usr/sbin/sysctl", "-w", "net.ipv4.ip_forward=1"], python_shell=False
        )


def test_sanitize_sysctl_value():
    assert (
        linux_sysctl._sanitize_sysctl_value("4096 131072  6291456")
        == "4096\t131072\t6291456"
    )


def test_sanitize_sysctl_value_int():
    assert linux_sysctl._sanitize_sysctl_value(1337) == "1337"


def test_persist_int(tmp_path):
    """
    Tests linux_sysctl.persist for an integer that is already set.
    """
    config = str(tmp_path / "sysctl.conf")
    config_file_content = "fs.suid_dumpable = 2\n"
    with fopen(config, "w", encoding="utf-8") as config_file:
        config_file.write(config_file_content)
    mock_run = MagicMock(return_value="2")
    mock_run_all = MagicMock()
    which_mock = MagicMock(return_value="/usr/sbin/sysctl")
    with patch("salt.utils.path.which", which_mock):
        with patch("os.path.exists", MagicMock(return_value=True)), patch.dict(
            linux_sysctl.__salt__, {"cmd.run": mock_run, "cmd.run_all": mock_run_all}
        ):
            assert (
                linux_sysctl.persist("fs.suid_dumpable", 2, config=config)
                == "Already set"
            )
            mock_run.assert_called_once_with(
                ["/usr/sbin/sysctl", "-n", "fs.suid_dumpable"], python_shell=False
            )
            mock_run_all.assert_not_called()
    assert os.path.isfile(config)
    with fopen(config, encoding="utf-8") as config_file:
        written = config_file.read()
    assert written == config_file_content


def test_persist_no_conf_failure():
    """
    Tests adding of config file failure
    """
    fopen_mock = MagicMock(side_effect=OSError())
    which_mock = MagicMock(return_value="/usr/sbin/sysctl")
    with patch("salt.utils.path.which", which_mock):
        with patch("os.path.isfile", MagicMock(return_value=False)), patch(
            "os.path.exists", MagicMock(return_value=False)
        ), patch("os.makedirs", MagicMock()), patch(
            "salt.utils.files.fopen", fopen_mock
        ):
            with pytest.raises(CommandExecutionError):
                linux_sysctl.persist("net.ipv4.ip_forward", 42, config=None)
    fopen_mock.assert_called_once()


def test_persist_no_conf_success():
    """
    Tests successful add of config file when previously not one
    """
    config = "/etc/sysctl.conf"
    with patch("os.path.isfile", MagicMock(return_value=False)), patch(
        "os.path.exists", MagicMock(return_value=True)
    ), patch("salt.utils.path.which", MagicMock(return_value="/bin/sysctl")):
        asn_cmd = {
            "pid": 1337,
            "retcode": 0,
            "stderr": "",
            "stdout": "net.ipv4.ip_forward = 1",
        }
        mock_asn_cmd = MagicMock(return_value=asn_cmd)

        sys_cmd = "systemd 208\n+PAM +LIBWRAP"
        mock_sys_cmd = MagicMock(return_value=sys_cmd)

        which_mock = MagicMock(return_value="/usr/sbin/sysctl")
        with patch("salt.utils.path.which", which_mock):
            with patch("salt.utils.files.fopen", mock_open()) as m_open, patch.dict(
                linux_sysctl.__context__, {"salt.utils.systemd.version": 232}
            ), patch.dict(
                linux_sysctl.__salt__,
                {"cmd.run_stdout": mock_sys_cmd, "cmd.run_all": mock_asn_cmd},
            ), patch.dict(
                systemd.__context__,
                {"salt.utils.systemd.booted": True, "salt.utils.systemd.version": 232},
            ):
                linux_sysctl.persist("net.ipv4.ip_forward", 1, config=config)
                writes = m_open.write_calls()
                assert writes == ["#\n# Kernel sysctl configuration\n#\n"], writes
        mock_asn_cmd.assert_called_once_with(
            ["/usr/sbin/sysctl", "-w", "net.ipv4.ip_forward=1"], python_shell=False
        )


def test_persist_read_conf_success():
    """
    Tests sysctl.conf read success
    """
    with patch("os.path.isfile", MagicMock(return_value=True)), patch(
        "os.path.exists", MagicMock(return_value=True)
    ):
        asn_cmd = {
            "pid": 1337,
            "retcode": 0,
            "stderr": "",
            "stdout": "net.ipv4.ip_forward = 1",
        }
        mock_asn_cmd = MagicMock(return_value=asn_cmd)

        sys_cmd = "systemd 208\n+PAM +LIBWRAP"
        mock_sys_cmd = MagicMock(return_value=sys_cmd)

        which_mock = MagicMock(return_value="/usr/sbin/sysctl")
        with patch("salt.utils.path.which", which_mock):
            with patch("salt.utils.files.fopen", mock_open()), patch.dict(
                linux_sysctl.__context__, {"salt.utils.systemd.version": 232}
            ), patch.dict(
                linux_sysctl.__salt__,
                {"cmd.run_stdout": mock_sys_cmd, "cmd.run_all": mock_asn_cmd},
            ), patch.dict(
                systemd.__context__, {"salt.utils.systemd.booted": True}
            ):
                assert linux_sysctl.persist("net.ipv4.ip_forward", 1) == "Updated"
        mock_asn_cmd.assert_called_once_with(
            ["/usr/sbin/sysctl", "-w", "net.ipv4.ip_forward=1"], python_shell=False
        )


def test_persist_parsing_file(tmp_path):
    """
    Tests linux_sysctl.persist to correctly parse the config file.
    """
    config = str(tmp_path / "sysctl.conf")
    with fopen(config, "w", encoding="utf-8") as config_file:
        config_file.write(
            """\
# Use dump-core from kdump-tools Debian package.
kernel.core_pattern = |/usr/share/kdump-tools/dump-core %p %s %t %e
 # Stop low-level messages on console = less logging
 kernel.printk  = 3 4 1 3

 net.ipv4.ip_forward=1
net.ipv4.tcp_rmem	=	4096	131072	6291456
"""
        )
    mock_run = MagicMock()
    mock_run_all = MagicMock(
        return_value={
            "pid": 1337,
            "retcode": 0,
            "stderr": "",
            "stdout": "net.ipv4.ip_forward = 0",
        }
    )

    which_mock = MagicMock(return_value="/usr/sbin/sysctl")
    with patch("salt.utils.path.which", which_mock):
        with patch("os.path.exists", MagicMock(return_value=True)), patch.dict(
            linux_sysctl.__salt__, {"cmd.run": mock_run, "cmd.run_all": mock_run_all}
        ):
            assert (
                linux_sysctl.persist("net.ipv4.ip_forward", "0", config=config)
                == "Updated"
            )
            mock_run.assert_not_called()
            mock_run_all.assert_called_once_with(
                ["/usr/sbin/sysctl", "-w", "net.ipv4.ip_forward=0"], python_shell=False
            )
    assert os.path.isfile(config)
    with fopen(config, encoding="utf-8") as config_file:
        written = config_file.read()
    assert (
        written
        == """\
# Use dump-core from kdump-tools Debian package.
kernel.core_pattern = |/usr/share/kdump-tools/dump-core %p %s %t %e
 # Stop low-level messages on console = less logging
 kernel.printk  = 3 4 1 3

net.ipv4.ip_forward = 0
net.ipv4.tcp_rmem	=	4096	131072	6291456
"""
    )


def test_persist_value_with_spaces_already_set(tmp_path):
    """
    Tests linux_sysctl.persist for a value with spaces that is already set.
    """
    config = str(tmp_path / "existing_sysctl_with_spaces.conf")
    value = "|/usr/share/kdump-tools/dump-core %p %s %t %e"
    config_file_content = f"kernel.core_pattern = {value}\n"
    with fopen(config, "w", encoding="utf-8") as config_file:
        config_file.write(config_file_content)
    mock_run = MagicMock(return_value=value)
    mock_run_all = MagicMock()
    which_mock = MagicMock(return_value="/usr/sbin/sysctl")
    with patch("salt.utils.path.which", which_mock):
        with patch("os.path.exists", MagicMock(return_value=True)), patch.dict(
            linux_sysctl.__salt__, {"cmd.run": mock_run, "cmd.run_all": mock_run_all}
        ):
            assert (
                linux_sysctl.persist("kernel.core_pattern", value, config=config)
                == "Already set"
            )
            mock_run.assert_called_once_with(
                ["/usr/sbin/sysctl", "-n", "kernel.core_pattern"], python_shell=False
            )
            mock_run_all.assert_not_called()
    assert os.path.isfile(config)
    with fopen(config, encoding="utf-8") as config_file:
        written = config_file.read()
    assert written == config_file_content


def test_persist_value_with_spaces_already_configured(tmp_path):
    """
    Tests linux_sysctl.persist for a value with spaces that is only configured.
    """
    config = str(tmp_path / "existing_sysctl_with_spaces.conf")
    value = "|/usr/share/kdump-tools/dump-core %p %s %t %e"
    config_file_content = f"kernel.core_pattern = {value}\n"
    with fopen(config, "w", encoding="utf-8") as config_file:
        config_file.write(config_file_content)
    mock_run = MagicMock(return_value="")
    mock_run_all = MagicMock(
        return_value={
            "pid": 1337,
            "retcode": 0,
            "stderr": "",
            "stdout": "kernel.core_pattern = " + value,
        }
    )
    which_mock = MagicMock(return_value="/usr/sbin/sysctl")
    with patch("salt.utils.path.which", which_mock):
        with patch("os.path.exists", MagicMock(return_value=True)), patch.dict(
            linux_sysctl.__salt__, {"cmd.run": mock_run, "cmd.run_all": mock_run_all}
        ):
            assert (
                linux_sysctl.persist("kernel.core_pattern", value, config=config)
                == "Updated"
            )
            mock_run.assert_called_once_with(
                ["/usr/sbin/sysctl", "-n", "kernel.core_pattern"], python_shell=False
            )
            mock_run_all.assert_called_once_with(
                ["/usr/sbin/sysctl", "-w", "kernel.core_pattern=" + value],
                python_shell=False,
            )
    assert os.path.isfile(config)
    with fopen(config, encoding="utf-8") as config_file:
        written = config_file.read()
    assert written == config_file_content


def test_persist_value_with_spaces_update_config(tmp_path):
    """
    Tests linux_sysctl.persist for a value with spaces that differs from the config.
    """
    config = str(tmp_path / "existing_sysctl_with_spaces.conf")
    value = "|/usr/share/kdump-tools/dump-core %p %s %t %e"
    with fopen(config, "w", encoding="utf-8") as config_file:
        config_file.write("kernel.core_pattern =\n")
    mock_run = MagicMock()
    mock_run_all = MagicMock(
        return_value={
            "pid": 1337,
            "retcode": 0,
            "stderr": "",
            "stdout": "kernel.core_pattern = " + value,
        }
    )
    which_mock = MagicMock(return_value="/usr/sbin/sysctl")
    with patch("salt.utils.path.which", which_mock):
        with patch("os.path.exists", MagicMock(return_value=True)), patch.dict(
            linux_sysctl.__salt__, {"cmd.run": mock_run, "cmd.run_all": mock_run_all}
        ):
            assert (
                linux_sysctl.persist("kernel.core_pattern", value, config=config)
                == "Updated"
            )
            mock_run.assert_not_called()
            mock_run_all.assert_called_once_with(
                ["/usr/sbin/sysctl", "-w", "kernel.core_pattern=" + value],
                python_shell=False,
            )
    assert os.path.isfile(config)
    with fopen(config, encoding="utf-8") as config_file:
        written = config_file.read()
    assert written == f"kernel.core_pattern = {value}\n"


def test_persist_value_with_spaces_new_file(tmp_path):
    """
    Tests linux_sysctl.persist for a value that contains spaces.
    """
    config = str(tmp_path / "sysctl_with_spaces.conf")
    value = "|/usr/share/kdump-tools/dump-core %p %s %t %e"
    mock_run_all = MagicMock(
        return_value={
            "pid": 1337,
            "retcode": 0,
            "stderr": "",
            "stdout": "kernel.core_pattern = " + value,
        }
    )
    which_mock = MagicMock(return_value="/usr/sbin/sysctl")
    with patch("salt.utils.path.which", which_mock):
        with patch("os.path.exists", MagicMock(return_value=True)), patch.dict(
            linux_sysctl.__salt__, {"cmd.run_all": mock_run_all}
        ):
            assert (
                linux_sysctl.persist("kernel.core_pattern", value, config=config)
                == "Updated"
            )
            mock_run_all.assert_called_once_with(
                ["/usr/sbin/sysctl", "-w", "kernel.core_pattern=" + value],
                python_shell=False,
            )
    assert os.path.isfile(config)
    with fopen(config, encoding="utf-8") as config_file:
        written = config_file.read()
    assert (
        written
        == """\
#
# Kernel sysctl configuration
#
kernel.core_pattern = |/usr/share/kdump-tools/dump-core %p %s %t %e
"""
    )


def test_persist_value_with_tabs_new_file(tmp_path):
    """
    Tests linux_sysctl.persist for a value that contains tabs.
    """
    config = str(tmp_path / "sysctl_with_tabs.conf")
    value = "|/usr/share/kdump-tools/dump-core\t%p\t%s\t%t\t%e"
    mock_run_all = MagicMock(
        return_value={
            "pid": 1337,
            "retcode": 0,
            "stderr": "",
            "stdout": "kernel.core_pattern = " + value,
        }
    )
    which_mock = MagicMock(return_value="/usr/sbin/sysctl")
    with patch("salt.utils.path.which", which_mock):
        with patch("os.path.exists", MagicMock(return_value=True)), patch.dict(
            linux_sysctl.__salt__, {"cmd.run_all": mock_run_all}
        ):
            assert (
                linux_sysctl.persist("kernel.core_pattern", value, config=config)
                == "Updated"
            )
            mock_run_all.assert_called_once_with(
                ["/usr/sbin/sysctl", "-w", "kernel.core_pattern=" + value],
                python_shell=False,
            )
    assert os.path.isfile(config)
    with fopen(config, encoding="utf-8") as config_file:
        written = config_file.read()
    assert (
        written
        == """\
#
# Kernel sysctl configuration
#
kernel.core_pattern = |/usr/share/kdump-tools/dump-core	%p	%s	%t	%e
"""
    )



import psutil
import pytest

import salt.modules.win_status as win_status

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


def test__get_connected_ips():
    # Let's find an active connection to test with
    port = None
    ip = None
    conns = psutil.net_connections()
    for conn in conns:
        if conn.status == psutil.CONN_ESTABLISHED:
            ip = conn.raddr.ip
            port = conn.raddr.port
            break
    assert port is not None
    assert ip is not None
    # Since this may return more than one IP, let's make sure our test IP is in
    # the list of IPs
    assert ip in win_status._get_connected_ips(port)



"""
NAPALM YANG state
=================

Manage the configuration of network devices according to
the YANG models (OpenConfig/IETF).

.. versionadded:: 2017.7.0

Dependencies
------------

- napalm-yang
- pyangbing > 0.5.11

To be able to load configuration on network devices,
it requires NAPALM_ library to be installed:  ``pip install napalm``.
Please check Installation_ for complete details.

.. _NAPALM: https://napalm.readthedocs.io
.. _Installation: https://napalm.readthedocs.io/en/latest/installation/index.html
"""

import logging

# Import salt modules
import salt.utils.files
import salt.utils.json
import salt.utils.napalm
import salt.utils.stringutils
import salt.utils.yaml

log = logging.getLogger(__file__)

try:
    # pylint: disable=unused-import
    import napalm_yang

    HAS_NAPALM_YANG = True
    # pylint: enable=unused-import
except ImportError:
    HAS_NAPALM_YANG = False


# ------------------------------------------------------------------------------
# state properties
# ------------------------------------------------------------------------------

__virtualname__ = "napalm_yang"

# ------------------------------------------------------------------------------
# global variables
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# property functions
# ------------------------------------------------------------------------------


def __virtual__():
    """
    NAPALM library must be installed for this module to work and run in a (proxy) minion.
    This module in particular requires also napalm-yang.
    """
    if not HAS_NAPALM_YANG:
        return (
            False,
            "Unable to load napalm_yang execution module: please install napalm-yang!",
        )
    return salt.utils.napalm.virtual(__opts__, __virtualname__, __file__)


# ------------------------------------------------------------------------------
# helper functions -- will not be exported
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# callable functions
# ------------------------------------------------------------------------------


def managed(name, data, **kwargs):
    """
    Manage the device configuration given the input data structured
    according to the YANG models.

    data
        YANG structured data.

    models
         A list of models to be used when generating the config.

    profiles: ``None``
        Use certain profiles to generate the config.
        If not specified, will use the platform default profile(s).

    compliance_report: ``False``
        Return the compliance report in the comment.

        .. versionadded:: 2017.7.3

    test: ``False``
        Dry run? If set as ``True``, will apply the config, discard
        and return the changes. Default: ``False`` and will commit
        the changes on the device.

    commit: ``True``
        Commit? Default: ``True``.

    debug: ``False``
        Debug mode. Will insert a new key under the output dictionary,
        as ``loaded_config`` containing the raw configuration loaded on the device.

    replace: ``False``
        Should replace the config with the new generate one?

    State SLS example:

    .. code-block:: jinja

        {%- set expected_config =  pillar.get('openconfig_interfaces_cfg') -%}
        interfaces_config:
          napalm_yang.managed:
            - data: {{ expected_config | json }}
            - models:
              - models.openconfig_interfaces
            - debug: true

    Pillar example:

    .. code-block:: yaml

        openconfig_interfaces_cfg:
          _kwargs:
            filter: true
          interfaces:
            interface:
              Et1:
                config:
                  mtu: 9000
              Et2:
                config:
                  description: "description example"
    """
    models = kwargs.get("models", None)
    if isinstance(models, tuple) and isinstance(models[0], list):
        models = models[0]
    ret = salt.utils.napalm.default_ret(name)
    test = kwargs.get("test", False) or __opts__.get("test", False)
    debug = kwargs.get("debug", False) or __opts__.get("debug", False)
    commit = kwargs.get("commit", True) or __opts__.get("commit", True)
    replace = kwargs.get("replace", False) or __opts__.get("replace", False)
    return_compliance_report = kwargs.get("compliance_report", False) or __opts__.get(
        "compliance_report", False
    )
    profiles = kwargs.get("profiles", [])
    temp_file = __salt__["temp.file"]()
    log.debug("Creating temp file: %s", temp_file)
    if "to_dict" not in data:
        data = {"to_dict": data}
    data = [data]
    with salt.utils.files.fopen(temp_file, "w") as file_handle:
        salt.utils.yaml.safe_dump(
            salt.utils.json.loads(salt.utils.json.dumps(data)),
            file_handle,
            encoding="utf-8",
        )
    device_config = __salt__["napalm_yang.parse"](
        *models, config=True, profiles=profiles
    )
    log.debug("Parsed the config from the device:")
    log.debug(device_config)
    compliance_report = __salt__["napalm_yang.compliance_report"](
        device_config, *models, filepath=temp_file
    )
    log.debug("Compliance report:")
    log.debug(compliance_report)
    complies = compliance_report.get("complies", False)
    if complies:
        ret.update({"result": True, "comment": "Already configured as required."})
        log.debug("All good here.")
        return ret
    log.debug("Does not comply, trying to generate and load config")
    data = data[0]["to_dict"]
    if "_kwargs" in data:
        data.pop("_kwargs")
    loaded_changes = __salt__["napalm_yang.load_config"](
        data,
        *models,
        profiles=profiles,
        test=test,
        debug=debug,
        commit=commit,
        replace=replace
    )
    log.debug("Loaded config result:")
    log.debug(loaded_changes)
    __salt__["file.remove"](temp_file)
    loaded_changes["compliance_report"] = compliance_report
    return salt.utils.napalm.loaded_ret(
        ret,
        loaded_changes,
        test,
        debug,
        opts=__opts__,
        compliance_report=return_compliance_report,
    )


def configured(name, data, **kwargs):
    """
    Configure the network device, given the input data strucuted
    according to the YANG models.

    .. note::
        The main difference between this function and ``managed``
        is that the later generates and loads the configuration
        only when there are differences between the existing
        configuration on the device and the expected
        configuration. Depending on the platform and hardware
        capabilities, one could be more optimal than the other.
        Additionally, the output of the ``managed`` is different,
        in such a way that the ``pchange`` field in the output
        contains structured data, rather than text.

    data
        YANG structured data.

    models
         A list of models to be used when generating the config.

    profiles: ``None``
        Use certain profiles to generate the config.
        If not specified, will use the platform default profile(s).

    test: ``False``
        Dry run? If set as ``True``, will apply the config, discard
        and return the changes. Default: ``False`` and will commit
        the changes on the device.

    commit: ``True``
        Commit? Default: ``True``.

    debug: ``False``
        Debug mode. Will insert a new key under the output dictionary,
        as ``loaded_config`` containing the raw configuration loaded on the device.

    replace: ``False``
        Should replace the config with the new generate one?

    State SLS example:

    .. code-block:: jinja

        {%- set expected_config =  pillar.get('openconfig_interfaces_cfg') -%}
        interfaces_config:
          napalm_yang.configured:
            - data: {{ expected_config | json }}
            - models:
              - models.openconfig_interfaces
            - debug: true

    Pillar example:

    .. code-block:: yaml

        openconfig_interfaces_cfg:
          _kwargs:
            filter: true
          interfaces:
            interface:
              Et1:
                config:
                  mtu: 9000
              Et2:
                config:
                  description: "description example"
    """
    models = kwargs.get("models", None)
    if isinstance(models, tuple) and isinstance(models[0], list):
        models = models[0]
    ret = salt.utils.napalm.default_ret(name)
    test = kwargs.get("test", False) or __opts__.get("test", False)
    debug = kwargs.get("debug", False) or __opts__.get("debug", False)
    commit = kwargs.get("commit", True) or __opts__.get("commit", True)
    replace = kwargs.get("replace", False) or __opts__.get("replace", False)
    profiles = kwargs.get("profiles", [])
    if "_kwargs" in data:
        data.pop("_kwargs")
    loaded_changes = __salt__["napalm_yang.load_config"](
        data,
        *models,
        profiles=profiles,
        test=test,
        debug=debug,
        commit=commit,
        replace=replace
    )
    return salt.utils.napalm.loaded_ret(ret, loaded_changes, test, debug)



import textwrap

import pytest

import salt.ext.tornado.ioloop
import salt.fileserver.gitfs as gitfs
import salt.utils.files
import salt.utils.gitfs
import salt.utils.platform
import salt.utils.win_functions
import salt.utils.yaml
from salt.utils.gitfs import GITPYTHON_MINVER, GITPYTHON_VERSION
from tests.support.mock import patch

try:
    # pylint: disable=unused-import
    import git

    # We still need to use GitPython here for temp repo setup, so we do need to
    # actually import it. But we don't need import pygit2 in this module, we
    # can just use the Version instances imported along with
    # salt.utils.gitfs to check if we have a compatible version.
    HAS_GITPYTHON = GITPYTHON_VERSION >= GITPYTHON_MINVER
except (ImportError, AttributeError):
    HAS_GITPYTHON = False


pytestmark = [
    pytest.mark.skipif(
        not HAS_GITPYTHON, reason=f"GitPython >= {GITPYTHON_MINVER} required"
    )
]


@pytest.fixture
def configure_loader_modules(tmp_path):
    opts = {
        "sock_dir": str(tmp_path / "sock_dir"),
        "gitfs_remotes": ["file://" + str(tmp_path / "repo_dir")],
        "cachedir": str(tmp_path / "cache_dir"),
        "gitfs_root": "",
        "fileserver_backend": ["gitfs"],
        "gitfs_base": "master",
        "gitfs_fallback": "",
        "fileserver_events": True,
        "transport": "zeromq",
        "gitfs_mountpoint": "",
        "gitfs_saltenv": [],
        "gitfs_saltenv_whitelist": [],
        "gitfs_saltenv_blacklist": [],
        "gitfs_user": "",
        "gitfs_password": "",
        "gitfs_insecure_auth": False,
        "gitfs_privkey": "",
        "gitfs_pubkey": "",
        "gitfs_passphrase": "",
        "gitfs_refspecs": [
            "+refs/heads/*:refs/remotes/origin/*",
            "+refs/tags/*:refs/tags/*",
        ],
        "gitfs_ssl_verify": True,
        "gitfs_disable_saltenv_mapping": False,
        "gitfs_ref_types": ["branch", "tag", "sha"],
        "gitfs_update_interval": 60,
        "__role": "master",
    }
    if salt.utils.platform.is_windows():
        opts["gitfs_remotes"][0] = opts["gitfs_remotes"][0].replace("\\", "/")

    return {gitfs: {"__opts__": opts}}


@pytest.fixture(scope="module", autouse=True)
def clear_instance_map():
    try:
        del salt.utils.gitfs.GitFS.instance_map[
            salt.ext.tornado.ioloop.IOLoop.current()
        ]
    except KeyError:
        pass


def test_per_saltenv_config():
    opts_override = textwrap.dedent(
        """
        gitfs_root: salt

        gitfs_saltenv:
          - baz:
            # when loaded, the "salt://" prefix will be removed
            - mountpoint: salt://baz_mountpoint
            - ref: baz_branch
            - root: baz_root

        gitfs_remotes:

          - file://{0}tmp/repo1:
            - saltenv:
              - foo:
                - ref: foo_branch
                - root: foo_root

          - file://{0}tmp/repo2:
            - mountpoint: repo2
            - saltenv:
              - baz:
                - mountpoint: abc
    """.format(
            "/" if salt.utils.platform.is_windows() else ""
        )
    )
    with patch.dict(gitfs.__opts__, salt.utils.yaml.safe_load(opts_override)):
        git_fs = salt.utils.gitfs.GitFS(
            gitfs.__opts__,
            gitfs.__opts__["gitfs_remotes"],
            per_remote_overrides=gitfs.PER_REMOTE_OVERRIDES,
            per_remote_only=gitfs.PER_REMOTE_ONLY,
        )

    # repo1 (branch: foo)
    # The mountpoint should take the default (from gitfs_mountpoint), while
    # ref and root should take the per-saltenv params.
    assert git_fs.remotes[0].mountpoint("foo") == ""
    assert git_fs.remotes[0].ref("foo") == "foo_branch"
    assert git_fs.remotes[0].root("foo") == "foo_root"

    # repo1 (branch: bar)
    # The 'bar' branch does not have a per-saltenv configuration set, so
    # each of the below values should fall back to global values.
    assert git_fs.remotes[0].mountpoint("bar") == ""
    assert git_fs.remotes[0].ref("bar") == "bar"
    assert git_fs.remotes[0].root("bar") == "salt"

    # repo1 (branch: baz)
    # The 'baz' branch does not have a per-saltenv configuration set, but
    # it is defined in the gitfs_saltenv parameter, so the values
    # from that parameter should be returned.
    assert git_fs.remotes[0].mountpoint("baz") == "baz_mountpoint"
    assert git_fs.remotes[0].ref("baz") == "baz_branch"
    assert git_fs.remotes[0].root("baz") == "baz_root"

    # repo2 (branch: foo)
    # The mountpoint should take the per-remote mountpoint value of
    # 'repo2', while ref and root should fall back to global values.
    assert git_fs.remotes[1].mountpoint("foo") == "repo2"
    assert git_fs.remotes[1].ref("foo") == "foo"
    assert git_fs.remotes[1].root("foo") == "salt"

    # repo2 (branch: bar)
    # The 'bar' branch does not have a per-saltenv configuration set, so
    # the mountpoint should take the per-remote mountpoint value of
    # 'repo2', while ref and root should fall back to global values.
    assert git_fs.remotes[1].mountpoint("bar") == "repo2"
    assert git_fs.remotes[1].ref("bar") == "bar"
    assert git_fs.remotes[1].root("bar") == "salt"

    # repo2 (branch: baz)
    # The 'baz' branch has the mountpoint configured as a per-saltenv
    # parameter. The other two should take the values defined in
    # gitfs_saltenv.
    assert git_fs.remotes[1].mountpoint("baz") == "abc"
    assert git_fs.remotes[1].ref("baz") == "baz_branch"
    assert git_fs.remotes[1].root("baz") == "baz_root"



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.fixture(scope="module")
def sys_mod(modules):
    return modules.sys


@pytest.fixture(scope="module")
def pillar(modules):
    return modules.pillar


def test_pillar_get_issue_61084(sys_mod):
    """
    Test issue 61084. `sys.argspec` should return valid data and not throw a
    TypeError due to pickling
    This should probably be a pre-commit check or something
    """
    result = sys_mod.argspec("pillar.get")
    assert isinstance(result, dict)
    assert isinstance(result.get("pillar.get"), dict)


def test_get_non_existing(pillar):
    """
    Tests pillar.get when the item does not exist. Should return an empty string
    """
    result = pillar.get("non-existing-pillar-item")
    assert result == ""


def test_get_default_none(pillar):
    """
    Tests pillar.get when default is set to `None`. Should return `None`
    """
    result = pillar.get("non-existing-pillar-item", default=None)
    assert result is None



"""
Generate baseline proxy minion grains for ESXi hosts.

.. Warning::
    This module will be deprecated in a future release of Salt. VMware strongly
    recommends using the
    `VMware Salt extensions <https://docs.saltproject.io/salt/extensions/salt-ext-modules-vmware/en/latest/all.html>`_
    instead of the ESXi module. Because the Salt extensions are newer and
    actively supported by VMware, they are more compatible with current versions
    of ESXi and they work well with the latest features in the VMware product
    line.


"""

import logging

import salt.utils.proxy
from salt.exceptions import SaltSystemExit

__proxyenabled__ = ["esxi"]
__virtualname__ = "esxi"

log = logging.getLogger(__file__)

GRAINS_CACHE = {}


def __virtual__():

    # import salt.utils.proxy again
    # so it is available for tests.
    import salt.utils.proxy

    try:
        if salt.utils.proxy.is_proxytype(__opts__, "esxi"):
            import salt.modules.vsphere

            return __virtualname__
    except KeyError:
        pass

    return False


def esxi():
    return _grains()


def kernel():
    return {"kernel": "proxy"}


def os():
    if not GRAINS_CACHE:
        GRAINS_CACHE.update(_grains())

    try:
        return {"os": GRAINS_CACHE.get("fullName")}
    except AttributeError:
        return {"os": "Unknown"}


def os_family():
    return {"os_family": "proxy"}


def _find_credentials(host):
    """
    Cycle through all the possible credentials and return the first one that
    works.
    """
    user_names = [__pillar__["proxy"].get("username", "root")]
    passwords = __pillar__["proxy"]["passwords"]
    for user in user_names:
        for password in passwords:
            try:
                # Try to authenticate with the given user/password combination
                ret = salt.modules.vsphere.system_info(
                    host=host, username=user, password=password
                )
            except SaltSystemExit:
                # If we can't authenticate, continue on to try the next password.
                continue
            # If we have data returned from above, we've successfully authenticated.
            if ret:
                return user, password
    # We've reached the end of the list without successfully authenticating.
    raise SaltSystemExit(
        "Cannot complete login due to an incorrect user name or password."
    )


def _grains():
    """
    Get the grains from the proxied device.
    """
    try:
        host = __pillar__["proxy"]["host"]
        if host:
            username, password = _find_credentials(host)
            protocol = __pillar__["proxy"].get("protocol")
            port = __pillar__["proxy"].get("port")
            ret = salt.modules.vsphere.system_info(
                host=host,
                username=username,
                password=password,
                protocol=protocol,
                port=port,
            )
            GRAINS_CACHE.update(ret)
    except KeyError:
        pass

    return GRAINS_CACHE



import datetime
import os
import sys

import pytest

import salt.modules.minion as minion
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {minion: {}}


# Test command fixtures
@pytest.fixture
def sample_restart_command():
    return ["/bin/some_command", "arg1", "arg2"]


@pytest.fixture
def daemon_argv():
    return ["/usr/bin/salt-minion", "-d"]


@pytest.fixture
def regular_argv():
    return ["/usr/bin/salt-minion"]


# Helper functions for common assertions
def assert_service_restart_called(mock_salt_dict, service_name, **kwargs):
    """Helper to assert service.restart was called with expected args"""
    mock_salt_dict["service.restart"].assert_called_once_with(service_name, **kwargs)


def assert_no_service_calls(mock_service_dict, mock_kill, mock_cmd_dict):
    """Helper to assert service.restart was not called"""
    mock_service_dict["service.restart"].assert_not_called()
    mock_kill.assert_called_once()


# Mock fixtures
@pytest.fixture
def mock_system_detection():
    """Mock functions to detect systemd and Windows systems"""
    with patch("salt.modules.minion._is_systemd_system") as mock_systemd, patch(
        "salt.modules.minion._is_windows_system"
    ) as mock_windows:
        yield mock_systemd, mock_windows


@pytest.fixture
def mock_minion_kill():
    """Mock the minion.kill function"""
    ret = {"retcode": 0, "killed": 1234}
    with patch("salt.modules.minion.kill", return_value=ret) as mock_kill:
        yield mock_kill


@pytest.fixture
def mock_salt_functions():
    """Mock salt functions used in minion.restart"""
    mocks = {
        "cmd.run_all": MagicMock(),
        "service.restart": MagicMock(),
        "config.get": MagicMock(return_value=False),
    }
    with patch.dict(minion.__salt__, mocks):
        yield mocks


def test_is_systemd_systemd():
    """Test if the system is a systemd system"""
    with patch.dict(minion.__grains__, {"kernel": "Linux"}):
        with patch("salt.utils.systemd.booted", return_value=True):
            assert minion._is_systemd_system() is True
        with patch("salt.utils.systemd.booted", return_value=False):
            assert minion._is_systemd_system() is False

    with patch.dict(minion.__grains__, {"kernel": "Windows"}):
        assert minion._is_systemd_system() is False


def test_is_windows_system():
    """Test if the system is a Windows system"""
    with patch.dict(minion.__grains__, {"kernel": "Windows"}):
        assert minion._is_windows_system() is True

    with patch.dict(minion.__grains__, {"kernel": "Linux"}):
        assert minion._is_windows_system() is False


def test_schedule_retry_systemd():
    """Test creating a scheduled retry for minion on Linux systems"""
    with patch("salt.modules.minion._is_systemd_system", return_value=False):
        assert minion._schedule_retry_systemd(60) is False
    with patch("salt.modules.minion._is_systemd_system", return_value=True):
        with patch("salt.utils.path.which", return_value="/usr/bin/systemd-run"):

            mock_cmd_run_all = MagicMock(return_value={"retcode": 0})
            with patch.dict(minion.__salt__, {"cmd.run_all": mock_cmd_run_all}):
                assert minion._schedule_retry_systemd(60)

                call_args = mock_cmd_run_all.call_args
                call_args_cmd = call_args[0][0]
                assert call_args_cmd[0] == "/usr/bin/systemd-run"
                assert call_args_cmd[1] == "--on-active=60"
                assert (
                    call_args_cmd[-1]
                    == "systemctl is-active salt-minion || systemctl start salt-minion"
                )


def test_schedule_retry_windows():
    """Test creating a scheduled retry for minion on Windows systems"""

    with patch("salt.modules.minion._is_windows_system", return_value=False):
        assert minion._schedule_retry_windows(60) is False

    fixed_time = datetime.datetime(2025, 1, 15, 14, 30, 0)
    expected_schedule_time = fixed_time + datetime.timedelta(seconds=60)

    with patch("salt.modules.minion._is_windows_system", return_value=True):
        with patch("salt.modules.minion.datetime") as mock_datetime:
            mock_datetime.datetime.now.return_value = fixed_time
            mock_datetime.timedelta = datetime.timedelta

            mock_create_task = MagicMock(return_value=True)

            with patch.dict(minion.__salt__, {"task.create_task": mock_create_task}):
                assert minion._schedule_retry(60)

                call_args = mock_create_task.call_args

                assert call_args.kwargs[
                    "start_date"
                ] == expected_schedule_time.strftime("%Y-%m-%d")
                assert call_args.kwargs[
                    "start_time"
                ] == expected_schedule_time.strftime("%H:%M:%S")
                assert call_args.kwargs["name"] == "retry-minion-restart"


def test_schedule_retry_invalid_delay():
    """Test _schedule_retry with invalid delay"""

    with pytest.raises(minion.CommandExecutionError):
        minion._schedule_retry("invalid_delay")


@pytest.mark.parametrize(
    "is_systemd,is_windows,expected_result,systemd_called,windows_called",
    [
        (False, False, False, False, False),  # Neither systemd nor windows
        (True, False, None, True, False),  # Systemd system
        (False, True, None, False, True),  # Windows system
    ],
    ids=["neither", "systemd", "windows"],
)
def test_schedule_retry_dispatch(
    is_systemd, is_windows, expected_result, systemd_called, windows_called
):
    """Test _schedule_retry dispatching"""

    mock_schedule_retry_systemd = MagicMock()
    mock_schedule_retry_windows = MagicMock()

    with patch(
        "salt.modules.minion._schedule_retry_systemd", mock_schedule_retry_systemd
    ), patch(
        "salt.modules.minion._schedule_retry_windows", mock_schedule_retry_windows
    ), patch(
        "salt.modules.minion._is_systemd_system", return_value=is_systemd
    ), patch(
        "salt.modules.minion._is_windows_system", return_value=is_windows
    ):
        result = minion._schedule_retry(60)

        # Check return value for the "neither" case
        if expected_result is not None:
            assert result is expected_result

        # Check mock calls
        if systemd_called:
            mock_schedule_retry_systemd.assert_called_once_with(60)
        else:
            mock_schedule_retry_systemd.assert_not_called()

        if windows_called:
            mock_schedule_retry_windows.assert_called_once_with(60)
        else:
            mock_schedule_retry_windows.assert_not_called()


# Tests of minion.restart()


def test_minion_restart_with_custom_command(
    mock_minion_kill, mock_salt_functions, sample_restart_command
):
    """Test behavior when minion_restart_command is configured"""
    mock_salt_functions["config.get"].return_value = sample_restart_command

    minion.restart()

    # Should call minion.kill()
    mock_minion_kill.assert_called_once()

    # Should call cmd.run_all with the custom command
    mock_salt_functions["cmd.run_all"].assert_called_once_with(
        sample_restart_command, env=os.environ
    )


@pytest.mark.parametrize(
    "is_systemd,is_windows,expected_service_call",
    [
        (True, False, {"service_name": "salt-minion", "kwargs": {"no_block": True}}),
        (False, True, {"service_name": "salt-minion", "kwargs": {}}),
    ],
)
def test_minion_restart_service_systems(
    is_systemd,
    is_windows,
    expected_service_call,
    mock_system_detection,
    mock_minion_kill,
    mock_salt_functions,
):
    """Test behavior on systemd and Windows systems"""

    mock_systemd, mock_windows = mock_system_detection
    mock_systemd.return_value = is_systemd
    mock_windows.return_value = is_windows

    minion.restart()

    # Should not call minion.kill
    mock_minion_kill.assert_not_called()

    # Should not call cmd.run_all
    mock_salt_functions["cmd.run_all"].assert_not_called()

    # Should call service.restart with expected parameters
    assert_service_restart_called(
        mock_salt_functions,
        expected_service_call["service_name"],
        **expected_service_call["kwargs"],
    )


@pytest.mark.parametrize(
    "systemd_override,win_service_override",
    [
        (False, None),  # systemd=False
        (None, False),  # win_service=False
    ],
)
def test_minion_restart_service_override(
    systemd_override,
    win_service_override,
    mock_system_detection,
    mock_minion_kill,
    mock_salt_functions,
    regular_argv,
):
    """Test behavior when service parameters are explicitly disabled"""

    mock_systemd, mock_windows = mock_system_detection
    mock_systemd.return_value = True if systemd_override is not None else False
    mock_windows.return_value = True if win_service_override is not None else False

    with patch.object(sys, "argv", regular_argv):
        kwargs = {}
        if systemd_override is not None:
            kwargs["systemd"] = systemd_override
        if win_service_override is not None:
            kwargs["win_service"] = win_service_override

        minion.restart(**kwargs)

        # Should use minion.kill instead of service.restart when overridden
        mock_minion_kill.assert_called_once()
        mock_salt_functions["service.restart"].assert_not_called()


@pytest.mark.parametrize(
    "argv,should_call_cmd_run",
    [
        (["/usr/bin/salt-minion", "-d"], True),
        (["/usr/bin/salt-minion"], False),
    ],
)
def test_minion_restart_non_service_systems(
    argv,
    should_call_cmd_run,
    mock_system_detection,
    mock_minion_kill,
    mock_salt_functions,
):
    """Test behavior on non-systemd/Windows systems"""

    mock_systemd, mock_windows = mock_system_detection
    mock_systemd.return_value = False
    mock_windows.return_value = False

    with patch.object(sys, "argv", argv):
        minion.restart()

        # Should always call kill on non systemd/Windows systems and not service.restart
        mock_minion_kill.assert_called_once()
        mock_salt_functions["service.restart"].assert_not_called()

        if should_call_cmd_run:
            mock_salt_functions["cmd.run_all"].assert_called_once_with(
                argv, env=os.environ
            )
        else:
            mock_salt_functions["cmd.run_all"].assert_not_called()


@pytest.mark.parametrize(
    "is_systemd,is_windows,retry_delay",
    [
        (True, False, None),  # systemd with default delay
        (True, False, 300),  # systemd with custom delay
        (False, True, None),  # Windows with default delay
        (False, True, 60),  # Windows with custom delay
    ],
)
def test_minion_restart_with_schedule_retry_success(
    is_systemd,
    is_windows,
    retry_delay,
    mock_system_detection,
    mock_salt_functions,
):
    """Test minion.restart with schedule_retry enabled - success case"""

    mock_systemd, mock_windows = mock_system_detection
    mock_systemd.return_value = is_systemd
    mock_windows.return_value = is_windows

    # Mock _schedule_retry to return success
    mock_schedule_retry_result = {"stderr": "success output"}
    with patch(
        "salt.modules.minion._schedule_retry", return_value=mock_schedule_retry_result
    ) as mock_schedule_retry:

        result = (
            minion.restart(schedule_retry=True, retry_delay=retry_delay)
            if retry_delay is not None
            else minion.restart(schedule_retry=True)
        )
        retry_delay = retry_delay if retry_delay is not None else 180

        # Verify _schedule_retry was called with correct delay
        mock_schedule_retry.assert_called_once_with(retry_delay)

        # Verify service.restart was called
        mock_salt_functions["service.restart"].assert_called_once()

        # Verify return structure contains schedule_retry info
        assert "service_restart" in result
        assert "schedule_retry" in result["service_restart"]
        assert result["service_restart"]["schedule_retry"]["delay"] == retry_delay

        # Check systemd-specific details
        if is_systemd:
            assert "detail" in result["service_restart"]["schedule_retry"]
            assert (
                result["service_restart"]["schedule_retry"]["detail"]
                == "success output"
            )

        # Verify comment mentions the scheduled retry
        assert "comment" in result
        expected_comment = (
            f"Scheduled retry for minion restart in {retry_delay} seconds"
        )
        assert expected_comment in result["comment"]

        # Verify successful restart
        assert result["retcode"] == 0


@pytest.mark.parametrize(
    "is_systemd,is_windows",
    [
        (True, False),  # systemd
        (False, True),  # Windows
    ],
)
def test_minion_restart_with_schedule_retry_failure(
    is_systemd,
    is_windows,
    mock_system_detection,
    mock_salt_functions,
):
    """Test minion.restart with schedule_retry enabled - failure case"""

    mock_systemd, mock_windows = mock_system_detection
    mock_systemd.return_value = is_systemd
    mock_windows.return_value = is_windows

    # Mock _schedule_retry to raise an exception
    with patch(
        "salt.modules.minion._schedule_retry",
        side_effect=minion.CommandExecutionError("Failed to add retry task"),
    ) as mock_schedule_retry:

        result = minion.restart(schedule_retry=True, retry_delay=120)

        # Verify _schedule_retry was called
        mock_schedule_retry.assert_called_once_with(120)

        # Verify service.restart was not called after schedule_retry failure
        mock_salt_functions["service.restart"].assert_not_called()

        # Verify error handling
        assert result["retcode"] != 0
        assert "service_restart" in result
        assert result["service_restart"]["result"] is False
        assert "stderr" in result["service_restart"]
        assert "Failed to add retry task" in result["service_restart"]["stderr"]

        # Verify error comment
        assert "comment" in result
        assert "Adding scheduled retry failed" in result["comment"]


def test_minion_restart_without_schedule_retry(
    mock_system_detection,
    mock_salt_functions,
):
    """Test minion.restart with schedule_retry disabled (default behavior)"""

    mock_systemd, mock_windows = mock_system_detection
    mock_systemd.return_value = True
    mock_windows.return_value = False

    with patch("salt.modules.minion._schedule_retry") as mock_schedule_retry:

        result = minion.restart(schedule_retry=False)

        # Verify _schedule_retry was not called
        mock_schedule_retry.assert_not_called()

        # Verify service.restart was called normally
        mock_salt_functions["service.restart"].assert_called_once()

        # Verify no schedule_retry info in result
        assert "service_restart" in result
        assert "schedule_retry" not in result["service_restart"]

        # Verify successful restart
        assert result["retcode"] == 0


@pytest.mark.parametrize(
    "is_systemd,is_windows",
    [
        (True, False),  # systemd
        (False, True),  # Windows
    ],
)
def test_minion_restart_schedule_retry_service_restart_failure(
    is_systemd,
    is_windows,
    mock_system_detection,
    mock_salt_functions,
):
    """Test minion.restart when schedule_retry succeeds but service.restart fails"""

    mock_systemd, mock_windows = mock_system_detection
    mock_systemd.return_value = is_systemd
    mock_windows.return_value = is_windows

    # Mock successful _schedule_retry
    mock_schedule_retry_result = {"stderr": "retry scheduled"}

    # Mock service.restart to fail
    mock_salt_functions["service.restart"].side_effect = minion.CommandExecutionError(
        "Service restart failed"
    )

    with patch(
        "salt.modules.minion._schedule_retry", return_value=mock_schedule_retry_result
    ) as mock_schedule_retry:

        result = minion.restart(schedule_retry=True, retry_delay=90)

        # Verify _schedule_retry was called successfully
        mock_schedule_retry.assert_called_once_with(90)

        # Verify service.restart was attempted
        mock_salt_functions["service.restart"].assert_called_once()

        # Verify schedule_retry info is still in result despite service failure
        assert "service_restart" in result
        assert "schedule_retry" in result["service_restart"]
        assert result["service_restart"]["schedule_retry"]["delay"] == 90

        # Verify service failure is handled
        assert result["retcode"] != 0
        assert result["service_restart"]["result"] is False
        assert "Service restart failed" in result["comment"]

        # Verify schedule retry comment is still present
        assert "Scheduled retry for minion restart in 90 seconds" in result["comment"]


def test_minion_restart_schedule_retry_non_service_system(
    mock_system_detection,
    mock_salt_functions,
    mock_minion_kill,
):
    """Test that schedule_retry is ignored on non-service systems"""

    mock_systemd, mock_windows = mock_system_detection
    mock_systemd.return_value = False
    mock_windows.return_value = False

    with patch("salt.modules.minion._schedule_retry") as mock_schedule_retry:
        with patch.object(sys, "argv", ["/usr/bin/salt-minion"]):

            result = minion.restart(schedule_retry=True, retry_delay=60)

            # Verify _schedule_retry was not called on non-service systems
            mock_schedule_retry.assert_not_called()

            # Verify kill was called instead
            mock_minion_kill.assert_called_once()

            # Verify no service_restart info
            assert "service_restart" not in result or not result["service_restart"]



import pytest

import salt.auth
import salt.runner

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.destructive_test,
    pytest.mark.windows_whitelisted,
]


@pytest.fixture
def client(client_config):
    return salt.runner.Runner(client_config)


def test_eauth(client, auth_creds):
    """
    Test executing master_call with lowdata

    The choice of using error.error for this is arbitrary and should be
    changed to some mocked function that is more testing friendly.
    """
    low = {"client": "runner", "fun": "error.error", **auth_creds}

    ret = client.master_call(**low)
    assert ret
    assert "jid" in ret
    assert "tag" in ret
    assert ret["tag"] == "salt/run/{}".format(ret["jid"])


def test_token(client, client_config, auth_creds):
    """
    Test executing master_call with lowdata

    The choice of using error.error for this is arbitrary and should be
    changed to some mocked function that is more testing friendly.
    """
    auth = salt.auth.LoadAuth(client_config)
    token = auth.mk_token(auth_creds)

    low = {
        "client": "runner",
        "fun": "error.error",
        "token": token["token"],
    }
    ret = client.master_call(**low)
    assert ret
    assert "jid" in ret
    assert "tag" in ret
    assert ret["tag"] == "salt/run/{}".format(ret["jid"])


def test_cmd_sync(client, auth_creds):
    low = {
        "client": "runner",
        "fun": "error.error",
        **auth_creds,
    }

    ret = client.cmd_sync(low.copy())
    assert ret == {}


def test_cmd_async(client, auth_creds):
    low = {
        "client": "runner",
        "fun": "error.error",
        **auth_creds,
    }

    ret = client.cmd_async(low.copy())
    assert ret
    assert "jid" in ret
    assert "tag" in ret
    assert ret["tag"] == "salt/run/{}".format(ret["jid"])


def test_cmd_sync_w_arg(client, auth_creds):
    low = {"fun": "test.arg", "foo": "Foo!", "bar": "Bar!", **auth_creds}

    ret = client.cmd_sync(low.copy())
    assert ret
    assert "kwargs" in ret
    assert ret["kwargs"]["foo"] == "Foo!"
    assert ret["kwargs"]["bar"] == "Bar!"


def test_wildcard_auth(client):
    # This test passes because of the following master config
    # external_auth:
    #   auto:
    #     '*':
    #       - '@wheel'
    #       - '@runner'
    #       - test.*
    low = {
        "fun": "test.arg",
        "foo": "Foo!",
        "bar": "Bar!",
        "username": "the_s0und_of_t3ch",
        "password": "willrockyou",
        "eauth": "auto",
    }

    ret = client.cmd_sync(low.copy())
    assert ret
    assert "kwargs" in ret
    assert ret["kwargs"]["foo"] == "Foo!"
    assert ret["kwargs"]["bar"] == "Bar!"


def test_full_return_kwarg(client, auth_creds):
    low = {"fun": "test.arg", **auth_creds}
    ret = client.cmd_sync(low.copy(), full_return=True)
    assert ret
    assert "data" in ret
    assert "success" in ret["data"]


def test_cmd_sync_arg_kwarg_parsing(client, auth_creds):
    low = {
        "client": "runner",
        "fun": "test.arg",
        "arg": ["foo", "bar=off", "baz={qux: 123}"],
        "kwarg": {"quux": "Quux"},
        "quuz": "on",
        **auth_creds,
    }

    ret = client.cmd_sync(low.copy())
    assert ret
    assert ret == {
        "args": ["foo"],
        "kwargs": {"bar": False, "baz": {"qux": 123}, "quux": "Quux", "quuz": "on"},
    }


def test_invalid_kwargs_are_ignored(client, auth_creds):
    low = {
        "client": "runner",
        "fun": "test.metasyntactic",
        "thiskwargisbad": "justpretendimnothere",
        **auth_creds,
    }

    ret = client.cmd_sync(low.copy())
    assert ret
    assert ret[0] == "foo"


def test_get_docs(client):
    ret = client.get_docs(arg="*")
    assert "auth.del_token" in ret
    assert "auth.mk_token" in ret
    assert "cache.clear_pillar" in ret
    assert "cache.grains" in ret
    assert "state.soft_kill" in ret
    assert "virt.start" in ret
    assert "test.arg" in ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import os

import pytest

import salt.modules.poudriere as poudriere
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {poudriere: {}}


def test_is_jail():
    """
    Test if it return True if jail exists False if not.
    """
    mock = MagicMock(return_value="salt stack")
    with patch.dict(poudriere.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
    ):
        assert poudriere.is_jail("salt")

        assert not poudriere.is_jail("SALT")


def test_make_pkgng_aware():
    """
    Test if it make jail ``jname`` pkgng aware.
    """
    temp_dir = os.path.join("tmp", "salt")
    conf_file = os.path.join("tmp", "salt", "salt-make.conf")
    ret1 = f"Could not create or find required directory {temp_dir}"
    ret2 = f"Looks like file {conf_file} could not be created"
    ret3 = {"changes": f"Created {conf_file}"}
    mock = MagicMock(return_value=temp_dir)
    mock_true = MagicMock(return_value=True)
    with patch.dict(
        poudriere.__salt__, {"config.option": mock, "file.write": mock_true}
    ):
        with patch.object(os.path, "isdir", MagicMock(return_value=False)):
            with patch.object(os, "makedirs", mock_true):
                assert poudriere.make_pkgng_aware("salt") == ret1

        with patch.object(os.path, "isdir", mock_true):
            assert poudriere.make_pkgng_aware("salt") == ret2

            with patch.object(os.path, "isfile", mock_true):
                assert poudriere.make_pkgng_aware("salt") == ret3


def test_parse_config():
    """
    Test if it returns a dict of poudriere main configuration definitions.
    """
    mock = MagicMock(return_value="/tmp/salt")
    with patch.dict(poudriere.__salt__, {"config.option": mock}), patch(
        "salt.utils.files.fopen", mock_open()
    ), patch.object(
        poudriere, "_check_config_exists", MagicMock(side_effect=[True, False])
    ):
        assert poudriere.parse_config() == {}

        assert poudriere.parse_config() == "Could not find /tmp/salt on file system"


def test_version():
    """
    Test if it return poudriere version.
    """
    mock = MagicMock(return_value="9.0-RELEASE")
    with patch.dict(poudriere.__salt__, {"cmd.run": mock}):
        assert poudriere.version() == "9.0-RELEASE"


def test_list_jails():
    """
    Test if it return a list of current jails managed by poudriere.
    """
    mock = MagicMock(return_value="salt stack")
    with patch.dict(poudriere.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
    ):
        assert poudriere.list_jails() == ["salt stack"]


def test_list_ports():
    """
    Test if it return a list of current port trees managed by poudriere.
    """
    mock = MagicMock(return_value="salt stack")
    with patch.dict(poudriere.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
    ):
        assert poudriere.list_ports() == ["salt stack"]


def test_create_jail():
    """
    Test if it creates a new poudriere jail if one does not exist.
    """
    mock_stack = MagicMock(return_value="90amd64 stack")
    mock_true = MagicMock(return_value=True)
    with patch.dict(poudriere.__salt__, {"cmd.run": mock_stack}), patch(
        "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
    ):
        assert poudriere.create_jail("90amd64", "amd64") == "90amd64 already exists"

        with patch.object(poudriere, "make_pkgng_aware", mock_true):
            assert (
                poudriere.create_jail("80amd64", "amd64")
                == "Issue creating jail 80amd64"
            )

    with patch.object(poudriere, "make_pkgng_aware", mock_true), patch(
        "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
    ):
        with patch.object(poudriere, "is_jail", MagicMock(side_effect=[False, True])):
            with patch.dict(poudriere.__salt__, {"cmd.run": mock_stack}):
                assert (
                    poudriere.create_jail("80amd64", "amd64") == "Created jail 80amd64"
                )


def test_update_jail():
    """
    Test if it run freebsd-update on `name` poudriere jail.
    """
    mock = MagicMock(return_value="90amd64 stack")
    with patch.dict(poudriere.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
    ):
        assert poudriere.update_jail("90amd64") == "90amd64 stack"

        assert poudriere.update_jail("80amd64") == "Could not find jail 80amd64"


def test_delete_jail():
    """
    Test if it deletes poudriere jail with `name`.
    """
    ret = "Looks like there was an issue deleting jail 90amd64"
    mock_stack = MagicMock(return_value="90amd64 stack")
    with patch.dict(poudriere.__salt__, {"cmd.run": mock_stack}), patch(
        "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
    ):
        assert poudriere.delete_jail("90amd64") == ret

        assert (
            poudriere.delete_jail("80amd64")
            == "Looks like jail 80amd64 has not been created"
        )

    ret1 = 'Deleted jail "80amd64" but was unable to remove jail make file'
    with patch.object(
        poudriere, "is_jail", MagicMock(side_effect=[True, False, True, False])
    ):
        with patch.dict(poudriere.__salt__, {"cmd.run": mock_stack}):
            with patch.object(
                poudriere, "_config_dir", MagicMock(return_value="/tmp/salt")
            ):
                assert poudriere.delete_jail("80amd64") == "Deleted jail 80amd64"

                with patch.object(os.path, "isfile", MagicMock(return_value=True)):
                    assert poudriere.delete_jail("80amd64") == ret1


def test_create_ports_tree():
    """
    Test if it not working need to run portfetch non interactive.
    """
    mock = MagicMock(return_value="salt stack")
    with patch.dict(poudriere.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
    ):
        assert poudriere.create_ports_tree() == "salt stack"


def test_update_ports_tree():
    """
    Test if it updates the ports tree, either the default
    or the `ports_tree` specified.
    """
    mock = MagicMock(return_value="salt stack")
    with patch.dict(poudriere.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
    ):
        assert poudriere.update_ports_tree("staging") == "salt stack"


def test_bulk_build():
    """
    Test if it run bulk build on poudriere server.
    """
    ret = "Could not find file /root/pkg_list on filesystem"
    mock = MagicMock(return_value="salt stack")
    with patch.dict(poudriere.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
    ):
        assert poudriere.bulk_build("90amd64", "/root/pkg_list") == ret

        with patch.object(os.path, "isfile", MagicMock(return_value=True)):
            assert (
                poudriere.bulk_build("90amd64", "/root/pkg_list")
                == "Could not find jail 90amd64"
            )

    ret = "There may have been an issue building packages dumping output: 90amd64 stack"
    with patch.object(os.path, "isfile", MagicMock(return_value=True)), patch(
        "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
    ):
        mock = MagicMock(return_value="90amd64 stack packages built")
        with patch.dict(poudriere.__salt__, {"cmd.run": mock}):
            assert (
                poudriere.bulk_build("90amd64", "/root/pkg_list")
                == "90amd64 stack packages built"
            )

        mock = MagicMock(return_value="90amd64 stack")
        with patch.dict(poudriere.__salt__, {"cmd.run": mock}):
            assert poudriere.bulk_build("90amd64", "/root/pkg_list") == ret


def test_info_jail():
    """
    Test to stdout the information poudriere jail with `name`.
    """
    ret = ["head-amd64"]
    mock_stack = MagicMock(return_value="head-amd64")
    with patch.dict(poudriere.__salt__, {"cmd.run": mock_stack}), patch(
        "salt.modules.poudriere._check_config_exists", MagicMock(return_value=True)
    ):
        assert poudriere.info_jail("head-amd64") == ret

        assert poudriere.info_jail("12-amd64") == "Could not find jail 12-amd64"

    ret1 = "Could not find jail 12-amd64"
    with patch.object(
        poudriere, "is_jail", MagicMock(side_effect=[True, False, True, False])
    ):
        with patch.dict(poudriere.__salt__, {"cmd.run": mock_stack}):
            with patch.object(
                poudriere, "_config_dir", MagicMock(return_value="/tmp/salt")
            ):
                assert poudriere.info_jail("12-amd64") == ["head-amd64"]

                with patch.object(os.path, "isfile", MagicMock(return_value=True)):
                    assert poudriere.info_jail("12-amd64") == ret1



"""
Integration tests for the mac_desktop execution module.
"""

import pytest

from salt.exceptions import CommandExecutionError

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_unless_on_darwin,
]


@pytest.fixture(scope="module")
def desktop(modules):
    return modules.desktop


def test_get_output_volume(desktop):
    """
    Tests the return of get_output_volume.
    """
    ret = desktop.get_output_volume()
    assert ret is not None


def test_set_output_volume(desktop):
    """
    Tests the return of set_output_volume.
    """
    current_vol = desktop.get_output_volume()
    if current_vol == "missing value":
        current_vol = 0
    try:
        to_set = 10
        if current_vol == str(to_set):
            to_set += 2
        new_vol = desktop.set_output_volume(str(to_set))
        check_vol = desktop.get_output_volume()
        assert new_vol == check_vol
    finally:
        # Set volume back to what it was before
        desktop.set_output_volume(current_vol)


def test_screensaver(desktop):
    """
    Tests the return of the screensaver function.
    """
    try:
        ret = desktop.screensaver()
    except CommandExecutionError as exc:
        pytest.skip("Skipping. Screensaver unavailable.")
    assert ret


def test_lock(desktop):
    """
    Tests the return of the lock function.
    """
    try:
        ret = desktop.lock()
    except CommandExecutionError as exc:
        pytest.skip("Skipping. Unable to lock screen.")
    assert ret


@pytest.mark.skipif(True, reason="Test is flaky. Is this really needed?")
def test_say(desktop):
    """
    Tests the return of the say function.
    """
    ret = desktop.say("hello", "world")
    assert ret



"""
The 'sysbench' module is used to analyze the
performance of the minions, right from the master!
It measures various system parameters such as
CPU, Memory, File I/O, Threads and Mutex.
"""

import re

import salt.utils.path


def __virtual__():
    """
    loads the module, if only sysbench is installed
    """
    # finding the path of the binary
    if salt.utils.path.which("sysbench"):
        return "sysbench"
    return (
        False,
        "The sysbench execution module failed to load: the sysbench binary is not in"
        " the path.",
    )


def _parser(result):
    """
    parses the output into a dictionary
    """

    # regexes to match
    _total_time = re.compile(r"total time:\s*(\d*.\d*s)")
    _total_execution = re.compile(r"event execution:\s*(\d*.\d*s?)")
    _min_response_time = re.compile(r"min:\s*(\d*.\d*ms)")
    _max_response_time = re.compile(r"max:\s*(\d*.\d*ms)")
    _avg_response_time = re.compile(r"avg:\s*(\d*.\d*ms)")
    _per_response_time = re.compile(r"95 percentile:\s*(\d*.\d*ms)")

    # extracting data
    total_time = re.search(_total_time, result).group(1)
    total_execution = re.search(_total_execution, result).group(1)
    min_response_time = re.search(_min_response_time, result).group(1)
    max_response_time = re.search(_max_response_time, result).group(1)
    avg_response_time = re.search(_avg_response_time, result).group(1)
    per_response_time = re.search(_per_response_time, result)
    if per_response_time is not None:
        per_response_time = per_response_time.group(1)

    # returning the data as dictionary
    return {
        "total time": total_time,
        "total execution time": total_execution,
        "minimum response time": min_response_time,
        "maximum response time": max_response_time,
        "average response time": avg_response_time,
        "95 percentile": per_response_time,
    }


def cpu():
    """
    Tests for the CPU performance of minions.

    CLI Examples:

    .. code-block:: bash

        salt '*' sysbench.cpu
    """

    # Test data
    max_primes = [500, 1000, 2500, 5000]

    # Initializing the test variables
    test_command = "sysbench --test=cpu --cpu-max-prime={0} run"
    result = None
    ret_val = {}

    # Test beings!
    for primes in max_primes:
        key = f"Prime numbers limit: {primes}"
        run_command = test_command.format(primes)
        result = __salt__["cmd.run"](run_command)
        ret_val[key] = _parser(result)

    return ret_val


def threads():
    """
    This tests the performance of the processor's scheduler

    CLI Example:

    .. code-block:: bash

        salt '*' sysbench.threads
    """

    # Test data
    thread_yields = [100, 200, 500, 1000]
    thread_locks = [2, 4, 8, 16]

    # Initializing the test variables
    test_command = "sysbench --num-threads=64 --test=threads "
    test_command += "--thread-yields={0} --thread-locks={1} run "
    result = None
    ret_val = {}

    # Test begins!
    for yields, locks in zip(thread_yields, thread_locks):
        key = f"Yields: {yields} Locks: {locks}"
        run_command = test_command.format(yields, locks)
        result = __salt__["cmd.run"](run_command)
        ret_val[key] = _parser(result)

    return ret_val


def mutex():
    """
    Tests the implementation of mutex

    CLI Examples:

    .. code-block:: bash

        salt '*' sysbench.mutex
    """

    # Test options and the values they take
    # --mutex-num = [50,500,1000]
    # --mutex-locks = [10000,25000,50000]
    # --mutex-loops = [2500,5000,10000]

    # Test data (Orthogonal test cases)
    mutex_num = [50, 50, 50, 500, 500, 500, 1000, 1000, 1000]
    locks = [10000, 25000, 50000, 10000, 25000, 50000, 10000, 25000, 50000]
    mutex_locks = []
    mutex_locks.extend(locks)
    mutex_loops = [2500, 5000, 10000, 10000, 2500, 5000, 5000, 10000, 2500]

    # Initializing the test variables
    test_command = "sysbench --num-threads=250 --test=mutex "
    test_command += "--mutex-num={0} --mutex-locks={1} --mutex-loops={2} run "
    result = None
    ret_val = {}

    # Test begins!
    for num, locks, loops in zip(mutex_num, mutex_locks, mutex_loops):
        key = f"Mutex: {num} Locks: {locks} Loops: {loops}"
        run_command = test_command.format(num, locks, loops)
        result = __salt__["cmd.run"](run_command)
        ret_val[key] = _parser(result)

    return ret_val


def memory():
    """
    This tests the memory for read and write operations.

    CLI Examples:

    .. code-block:: bash

        salt '*' sysbench.memory
    """

    # test defaults
    # --memory-block-size = 10M
    # --memory-total-size = 1G

    # We test memory read / write against global / local scope of memory
    # Test data
    memory_oper = ["read", "write"]
    memory_scope = ["local", "global"]

    # Initializing the test variables
    test_command = "sysbench --num-threads=64 --test=memory "
    test_command += "--memory-oper={0} --memory-scope={1} "
    test_command += "--memory-block-size=1K --memory-total-size=32G run "
    result = None
    ret_val = {}

    # Test begins!
    for oper in memory_oper:
        for scope in memory_scope:
            key = f"Operation: {oper} Scope: {scope}"
            run_command = test_command.format(oper, scope)
            result = __salt__["cmd.run"](run_command)
            ret_val[key] = _parser(result)

    return ret_val


def fileio():
    """
    This tests for the file read and write operations
    Various modes of operations are

    * sequential write
    * sequential rewrite
    * sequential read
    * random read
    * random write
    * random read and write

    The test works with 32 files with each file being 1Gb in size
    The test consumes a lot of time. Be patient!

    CLI Examples:

    .. code-block:: bash

        salt '*' sysbench.fileio
    """

    # Test data
    test_modes = ["seqwr", "seqrewr", "seqrd", "rndrd", "rndwr", "rndrw"]

    # Initializing the required variables
    test_command = "sysbench --num-threads=16 --test=fileio "
    test_command += "--file-num=32 --file-total-size=1G --file-test-mode={0} "
    result = None
    ret_val = {}

    # Test begins!
    for mode in test_modes:
        key = f"Mode: {mode}"

        # Prepare phase
        run_command = (test_command + "prepare").format(mode)
        __salt__["cmd.run"](run_command)

        # Test phase
        run_command = (test_command + "run").format(mode)
        result = __salt__["cmd.run"](run_command)
        ret_val[key] = _parser(result)

        # Clean up phase
        run_command = (test_command + "cleanup").format(mode)
        __salt__["cmd.run"](run_command)

    return ret_val


def ping():

    return True



"""
Integration tests for the vault modules
"""

import json
import logging
import subprocess
import time

import pytest
from pytestshellutils.utils.processes import ProcessResult

import salt.utils.path
from tests.support.helpers import PatchedEnviron
from tests.support.runtests import RUNTIME_VARS

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd", "vault", "getent"),
]


@pytest.fixture(scope="module")
def patched_environ(vault_port):
    with PatchedEnviron(VAULT_ADDR=f"http://127.0.0.1:{vault_port}"):
        yield


def vault_container_version_id(value):
    return f"vault=={value}"


@pytest.fixture(
    scope="module",
    autouse=True,
    params=["0.9.6", "1.3.1", "latest"],
    ids=vault_container_version_id,
)
def vault_container_version(request, salt_factories, vault_port, patched_environ):
    vault_version = request.param
    vault_binary = salt.utils.path.which("vault")
    config = {
        "backend": {"file": {"path": "/vault/file"}},
        "default_lease_ttl": "168h",
        "max_lease_ttl": "720h",
    }
    factory = salt_factories.get_container(
        "vault",
        f"ghcr.io/saltstack/salt-ci-containers/vault:{vault_version}",
        check_ports=[vault_port],
        container_run_kwargs={
            "ports": {"8200/tcp": vault_port},
            "environment": {
                "VAULT_DEV_ROOT_TOKEN_ID": "testsecret",
                "VAULT_LOCAL_CONFIG": json.dumps(config),
            },
            "cap_add": ["IPC_LOCK"],
        },
        pull_before_start=True,
        skip_on_pull_failure=True,
        skip_if_docker_client_not_connectable=True,
    )
    with factory.started() as factory:
        attempts = 0
        while attempts < 3:
            attempts += 1
            time.sleep(1)
            proc = subprocess.run(
                [vault_binary, "login", "token=testsecret"],
                check=False,
                capture_output=True,
                text=True,
            )
            if proc.returncode == 0:
                break
            ret = ProcessResult(
                returncode=proc.returncode,
                stdout=proc.stdout,
                stderr=proc.stderr,
                cmdline=proc.args,
            )
            log.debug("Failed to authenticate against vault:\n%s", ret)
            time.sleep(4)
        else:
            pytest.fail("Failed to login to vault")

        proc = subprocess.run(
            [
                vault_binary,
                "policy",
                "write",
                "testpolicy",
                f"{RUNTIME_VARS.FILES}/vault.hcl",
            ],
            check=False,
            capture_output=True,
            text=True,
        )
        if proc.returncode != 0:
            ret = ProcessResult(
                returncode=proc.returncode,
                stdout=proc.stdout,
                stderr=proc.stderr,
                cmdline=proc.args,
            )
            log.debug("Failed to assign policy to vault:\n%s", ret)
            pytest.fail("unable to assign policy to vault")
        if vault_version in ("1.3.1", "latest"):
            proc = subprocess.run(
                [vault_binary, "secrets", "enable", "kv-v2"],
                check=False,
                capture_output=True,
                text=True,
            )
            ret = ProcessResult(
                returncode=proc.returncode,
                stdout=proc.stdout,
                stderr=proc.stderr,
                cmdline=proc.args,
            )
            if proc.returncode != 0:
                log.debug("Failed to enable kv-v2:\n%s", ret)
                pytest.fail("Could not enable kv-v2")

            if "path is already in use at kv-v2/" in proc.stdout:
                pass
            elif "Success" in proc.stdout:
                pass
            else:
                log.debug("Failed to enable kv-v2:\n%s", ret)
                pytest.fail(f"Could not enable kv-v2 {proc.stdout}")
            if vault_version == "latest":
                proc = subprocess.run(
                    [
                        vault_binary,
                        "secrets",
                        "enable",
                        "-version=2",
                        "-path=salt/",
                        "kv",
                    ],
                    check=False,
                    capture_output=True,
                    text=True,
                )
                ret = ProcessResult(
                    returncode=proc.returncode,
                    stdout=proc.stdout,
                    stderr=proc.stderr,
                    cmdline=proc.args,
                )
                if proc.returncode != 0:
                    log.debug("Failed to enable kv-v2:\n%s", ret)
                    pytest.fail("Could not enable kv-v2")

                if "path is already in use at kv-v2/" in proc.stdout:
                    pass
                elif "Success" in proc.stdout:
                    proc = subprocess.run(
                        [
                            vault_binary,
                            "kv",
                            "put",
                            "salt/user1",
                            "password=p4ssw0rd",
                            "desc=test user",
                        ],
                        check=False,
                        capture_output=True,
                        text=True,
                    )
                    ret = ProcessResult(
                        returncode=proc.returncode,
                        stdout=proc.stdout,
                        stderr=proc.stderr,
                        cmdline=proc.args,
                    )
                    if proc.returncode != 0:
                        log.debug("Failed to enable kv-v2:\n%s", ret)
                        pytest.fail("Could not enable kv-v2")
                    if "path is already in use at kv-v2/" in proc.stdout:
                        pass
                    elif "created_time" in proc.stdout:
                        proc = subprocess.run(
                            [
                                vault_binary,
                                "kv",
                                "put",
                                "salt/user/user1",
                                "password=p4ssw0rd",
                                "desc=test user",
                            ],
                            check=False,
                            capture_output=True,
                            text=True,
                        )
                        ret = ProcessResult(
                            returncode=proc.returncode,
                            stdout=proc.stdout,
                            stderr=proc.stderr,
                            cmdline=proc.args,
                        )
                        if proc.returncode != 0:
                            log.debug("Failed to enable kv-v2:\n%s", ret)
                            pytest.fail("Could not enable kv-v2")

                        if "path is already in use at kv-v2/" in proc.stdout:
                            pass
                        elif "created_time" in proc.stdout:
                            proc = subprocess.run(
                                [vault_binary, "kv", "get", "salt/user1"],
                                check=False,
                                capture_output=True,
                                text=True,
                            )
                            ret = ProcessResult(
                                returncode=proc.returncode,
                                stdout=proc.stdout,
                                stderr=proc.stderr,
                                cmdline=proc.args,
                            )

                else:
                    log.debug("Failed to enable kv-v2:\n%s", ret)
                    pytest.fail(f"Could not enable kv-v2 {proc.stdout}")
        yield vault_version


def test_sdb(salt_call_cli):
    ret = salt_call_cli.run(
        "sdb.set", uri="sdb://sdbvault/secret/test/test_sdb/foo", value="bar"
    )
    assert ret.returncode == 0
    assert ret.data is True
    ret = salt_call_cli.run("sdb.get", uri="sdb://sdbvault/secret/test/test_sdb/foo")
    assert ret.returncode == 0
    assert ret.data
    assert ret.data == "bar"


def test_sdb_runner(salt_run_cli):
    ret = salt_run_cli.run(
        "sdb.set", uri="sdb://sdbvault/secret/test/test_sdb_runner/foo", value="bar"
    )
    assert ret.returncode == 0
    assert ret.data is True
    ret = salt_run_cli.run(
        "sdb.get", uri="sdb://sdbvault/secret/test/test_sdb_runner/foo"
    )
    assert ret.returncode == 0
    assert ret.stdout
    assert ret.stdout == "bar"


def test_config(salt_call_cli, pillar_tree):
    ret = salt_call_cli.run(
        "sdb.set", uri="sdb://sdbvault/secret/test/test_pillar_sdb/foo", value="bar"
    )
    assert ret.returncode == 0
    assert ret.data is True
    ret = salt_call_cli.run("config.get", "test_vault_pillar_sdb")
    assert ret.returncode == 0
    assert ret.data
    assert ret.data == "bar"


def test_sdb_kv2_kvv2_path_local(salt_call_cli, vault_container_version):
    if vault_container_version not in ["1.3.1", "latest"]:
        pytest.skip(f"Test not applicable to vault {vault_container_version}")

    ret = salt_call_cli.run(
        "sdb.set", uri="sdb://sdbvault/kv-v2/test/test_sdb/foo", value="bar"
    )
    assert ret.returncode == 0
    assert ret.data is True
    ret = salt_call_cli.run(
        "--local", "sdb.get", "sdb://sdbvault/kv-v2/test/test_sdb/foo"
    )
    assert ret.data
    assert ret.data == "bar"


def test_sdb_kv_dual_item(salt_call_cli, vault_container_version):
    if vault_container_version not in ["latest"]:
        pytest.skip(f"Test not applicable to vault {vault_container_version}")
    ret = salt_call_cli.run("--local", "sdb.get", "sdb://sdbvault/salt/data/user1")
    assert ret.data
    assert ret.data == {"desc": "test user", "password": "p4ssw0rd"}



"""
Python pretty-print (pprint)
============================

The python pretty-print system was once the default outputter. It simply
passes the return data through to ``pprint.pformat`` and prints the results.

CLI Example:

.. code-block:: bash

    salt '*' foo.bar --out=pprint

Example output:

.. code-block:: python

    {'saltmine': {'foo': {'bar': 'baz',
                          'dictionary': {'abc': 123, 'def': 456},
                          'list': ['Hello', 'World']}}}
"""

import pprint

# Define the module's virtual name
__virtualname__ = "pprint"


def __virtual__():
    """
    Change the name to pprint
    """
    return __virtualname__


def output(data, **kwargs):  # pylint: disable=unused-argument
    """
    Print out via pretty print
    """
    if isinstance(data, Exception):
        data = str(data)
    if "output_indent" in __opts__ and __opts__["output_indent"] >= 0:
        return pprint.pformat(data, indent=__opts__["output_indent"])
    return pprint.pformat(data)



"""
:maintainer: Jack Kuan <kjkuan@gmail.com>
:maturity: new
:platform: all

A Python DSL for generating Salt's highstate data structure.

This module is intended to be used with the `pydsl` renderer,
but can also be used on its own. Here's what you can do with
Salt PyDSL::

    # Example translated from the online salt tutorial

    apache = state('apache')
    apache.pkg.installed()
    apache.service.running() \\
                  .watch(pkg='apache',
                         file='/etc/httpd/conf/httpd.conf',
                         user='apache')

    if __grains__['os'] == 'RedHat':
        apache.pkg.installed(name='httpd')
        apache.service.running(name='httpd')

    apache.group.present(gid=87).require(apache.pkg)
    apache.user.present(uid=87, gid=87,
                        home='/var/www/html',
                        shell='/bin/nologin') \\
               .require(apache.group)

    state('/etc/httpd/conf/httpd.conf').file.managed(
        source='salt://apache/httpd.conf',
        user='root',
        group='root',
        mode=644)


Example with ``include`` and ``extend``, translated from
the online salt tutorial::

    include('http', 'ssh')
    extend(
        state('apache').file(
            name='/etc/httpd/conf/httpd.conf',
            source='salt://http/httpd2.conf'
        ),
        state('ssh-server').service.watch(file='/etc/ssh/banner')
    )
    state('/etc/ssh/banner').file.managed(source='salt://ssh/banner')


Example of a ``cmd`` state calling a python function::

    def hello(s):
        s = "hello world, %s" % s
        return dict(result=True, changes=dict(changed=True, output=s))

    state('hello').cmd.call(hello, 'pydsl!')

"""

# Implementation note:
#  - There's a bit of terminology mix-up here:
#    - what I called a state or state declaration here is actually
#      an ID declaration.
#    - what I called a module or a state module actually corresponds
#      to a state declaration.
#    - and a state function is a function declaration.


# TODOs:
#
#  - support exclude declarations
#
#  - allow this:
#      state('X').cmd.run.cwd = '/'
#      assert state('X').cmd.run.cwd == '/'
#
#  - make it possible to remove:
#    - state declarations
#    - state module declarations
#    - state func and args
#


from collections import OrderedDict
from uuid import uuid4 as _uuid

from salt.state import HighState

REQUISITES = set(
    "listen require watch prereq use listen_in require_in watch_in prereq_in use_in"
    " onchanges onfail".split()
)


class PyDslError(Exception):
    pass


class Options(dict):
    def __getattr__(self, name):
        return self.get(name)


SLS_MATCHES = None


class Sls:
    def __init__(self, sls, saltenv, rendered_sls):
        self.name = sls
        self.saltenv = saltenv
        self.includes = []
        self.included_highstate = HighState.get_active().building_highstate
        self.extends = []
        self.decls = []
        self.options = Options()
        self.funcs = []  # track the ordering of state func declarations
        self.rendered_sls = rendered_sls  # a set of names of rendered sls modules

        if not HighState.get_active():
            raise PyDslError("PyDSL only works with a running high state!")

    @classmethod
    def get_all_decls(cls):
        return HighState.get_active()._pydsl_all_decls

    @classmethod
    def get_render_stack(cls):
        return HighState.get_active()._pydsl_render_stack

    def set(self, **options):
        self.options.update(options)

    def include(self, *sls_names, **kws):
        if "env" in kws:
            # "env" is not supported; Use "saltenv".
            kws.pop("env")

        saltenv = kws.get("saltenv", self.saltenv)

        if kws.get("delayed", False):
            for incl in sls_names:
                self.includes.append((saltenv, incl))
            return

        HIGHSTATE = HighState.get_active()

        global SLS_MATCHES
        if SLS_MATCHES is None:
            SLS_MATCHES = HIGHSTATE.top_matches(HIGHSTATE.get_top())

        highstate = self.included_highstate
        slsmods = []  # a list of pydsl sls modules rendered.
        for sls in sls_names:
            r_env = f"{saltenv}:{sls}"
            if r_env not in self.rendered_sls:
                self.rendered_sls.add(
                    sls
                )  # needed in case the starting sls uses the pydsl renderer.
                histates, errors = HIGHSTATE.render_state(
                    sls, saltenv, self.rendered_sls, SLS_MATCHES
                )
                HIGHSTATE.merge_included_states(highstate, histates, errors)
                if errors:
                    raise PyDslError("\n".join(errors))
                HIGHSTATE.clean_duplicate_extends(highstate)

            state_id = f"_slsmod_{sls}"
            if state_id not in highstate:
                slsmods.append(None)
            else:
                for arg in highstate[state_id]["stateconf"]:
                    if isinstance(arg, dict) and next(iter(arg)) == "slsmod":
                        slsmods.append(arg["slsmod"])
                        break

        if not slsmods:
            return None
        return slsmods[0] if len(slsmods) == 1 else slsmods

    def extend(self, *state_funcs):
        if self.options.ordered or self.last_func():
            raise PyDslError("Cannot extend() after the ordered option was turned on!")
        for f in state_funcs:
            state_id = f.mod._state_id
            self.extends.append(self.get_all_decls().pop(state_id))
            i = len(self.decls)
            for decl in reversed(self.decls):
                i -= 1
                if decl._id == state_id:
                    del self.decls[i]
                    break

    def state(self, id=None):
        if not id:
            id = f".{_uuid()}"
            # adds a leading dot to make use of stateconf's namespace feature.
        try:
            return self.get_all_decls()[id]
        except KeyError:
            self.get_all_decls()[id] = s = StateDeclaration(id)
            self.decls.append(s)
            return s

    def last_func(self):
        return self.funcs[-1] if self.funcs else None

    def track_func(self, statefunc):
        self.funcs.append(statefunc)

    def to_highstate(self, slsmod):
        # generate a state that uses the stateconf.set state, which
        # is a no-op state, to hold a reference to the sls module
        # containing the DSL statements. This is to prevent the module
        # from being GC'ed, so that objects defined in it will be
        # available while salt is executing the states.
        slsmod_id = "_slsmod_" + self.name
        self.state(slsmod_id).stateconf.set(slsmod=slsmod)
        del self.get_all_decls()[slsmod_id]

        highstate = OrderedDict()
        if self.includes:
            highstate["include"] = [{t[0]: t[1]} for t in self.includes]
        if self.extends:
            highstate["extend"] = extend = OrderedDict()
            for ext in self.extends:
                extend[ext._id] = ext._repr(context="extend")
        for decl in self.decls:
            highstate[decl._id] = decl._repr()

        if self.included_highstate:
            errors = []
            HighState.get_active().merge_included_states(
                highstate, self.included_highstate, errors
            )
            if errors:
                raise PyDslError("\n".join(errors))
        return highstate

    def load_highstate(self, highstate):
        for sid, decl in highstate.items():
            s = self.state(sid)
            for modname, args in decl.items():
                if "." in modname:
                    modname, funcname = modname.rsplit(".", 1)
                else:
                    funcname = next(x for x in args if isinstance(x, str))
                    args.remove(funcname)
                mod = getattr(s, modname)
                named_args = {}
                for x in args:
                    if isinstance(x, dict):
                        k, v = next(iter(x.items()))
                        named_args[k] = v
                mod(funcname, **named_args)


class StateDeclaration:
    def __init__(self, id):
        self._id = id
        self._mods = []

    def __getattr__(self, name):
        for m in self._mods:
            if m._name == name:
                return m
        m = StateModule(name, self._id)
        self._mods.append(m)
        return m

    __getitem__ = __getattr__

    def __str__(self):
        return self._id

    def __iter__(self):
        return iter(self._mods)

    def _repr(self, context=None):
        return OrderedDict(m._repr(context) for m in self)

    def __call__(self, check=True):
        sls = Sls.get_render_stack()[-1]
        if self._id in sls.get_all_decls():
            last_func = sls.last_func()
            if last_func and self._mods[-1]._func is not last_func:
                raise PyDslError(
                    "Cannot run state({}: {}) that is required by a runtime "
                    "state({}: {}), at compile time.".format(
                        self._mods[-1]._name,
                        self._id,
                        last_func.mod,
                        last_func.mod._state_id,
                    )
                )
            sls.get_all_decls().pop(self._id)
            sls.decls.remove(self)
            self._mods[0]._func._remove_auto_require()
            for m in self._mods:
                try:
                    sls.funcs.remove(m._func)
                except ValueError:
                    pass

        result = HighState.get_active().state.functions["state.high"](
            {self._id: self._repr()}
        )

        if not isinstance(result, dict):
            # A list is an error
            raise PyDslError(
                "An error occurred while running highstate: {}".format(
                    "; ".join(result)
                )
            )

        result = sorted(result.items(), key=lambda t: t[1]["__run_num__"])
        if check:
            for k, v in result:
                if not v["result"]:
                    import pprint

                    raise PyDslError(
                        "Failed executing low state at compile time:\n{}".format(
                            pprint.pformat({k: v})
                        )
                    )
        return result


class StateModule:
    def __init__(self, name, parent_decl):
        self._state_id = parent_decl
        self._name = name
        self._func = None

    def __getattr__(self, name):
        if self._func:
            if name == self._func.name:
                return self._func
            else:
                if name not in REQUISITES:
                    if self._func.name:
                        raise PyDslError(
                            "Multiple state functions({}) not allowed in a "
                            "state module({})!".format(name, self._name)
                        )
                    self._func.name = name
                    return self._func
                return getattr(self._func, name)

        if name in REQUISITES:
            self._func = f = StateFunction(None, self)
            return getattr(f, name)
        else:
            self._func = f = StateFunction(name, self)
            return f

    def __call__(self, _fname, *args, **kws):
        return getattr(self, _fname).configure(args, kws)

    def __str__(self):
        return self._name

    def _repr(self, context=None):
        return (self._name, self._func._repr(context))


def _generate_requsite_method(t):
    def req(self, *args, **kws):
        for mod in args:
            self.reference(t, mod, None)
        for mod_ref in kws.items():
            self.reference(t, *mod_ref)
        return self

    return req


class StateFunction:
    def __init__(self, name, parent_mod):
        self.mod = parent_mod
        self.name = name
        self.args = []

        # track the position of the auto-added require for easy
        # removal if run at compile time.
        self.require_index = None

        sls = Sls.get_render_stack()[-1]
        if sls.options.ordered:
            last_f = sls.last_func()
            if last_f:
                self.require(last_f.mod)
                self.require_index = len(self.args) - 1
            sls.track_func(self)

    def _remove_auto_require(self):
        if self.require_index is not None:
            del self.args[self.require_index]
            self.require_index = None

    def __call__(self, *args, **kws):
        self.configure(args, kws)
        return self

    def _repr(self, context=None):
        if not self.name and context != "extend":
            raise PyDslError(
                f"No state function specified for module: {self.mod._name}"
            )
        if not self.name and context == "extend":
            return self.args
        return [self.name] + self.args

    def configure(self, args, kws):
        args = list(args)
        if args:
            first = args[0]
            if (
                self.mod._name == "cmd"
                and self.name in ("call", "wait_call")
                and callable(first)
            ):

                args[0] = first.__name__
                kws = dict(func=first, args=args[1:], kws=kws)
                del args[1:]

            args[0] = dict(name=args[0])

        for k, v in kws.items():
            args.append({k: v})

        self.args.extend(args)
        return self

    def reference(self, req_type, mod, ref):
        if isinstance(mod, StateModule):
            ref = mod._state_id
        elif not (mod and ref):
            raise PyDslError(f"Invalid a requisite reference declaration! {mod}: {ref}")
        self.args.append({req_type: [{str(mod): str(ref)}]})

    ns = locals()
    for req_type in REQUISITES:
        ns[req_type] = _generate_requsite_method(req_type)
    del ns
    del req_type



"""
Manage vms running on the OpenBSD VMM hypervisor using vmctl(8).

.. versionadded:: 2019.2.0

:codeauthor: ``Jasper Lievisse Adriaanse <jasper@openbsd.org>``

.. note::

    This module requires the `vmd` service to be running on the OpenBSD
    target machine.
"""

import logging
import re

import salt.utils.path
from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only works on OpenBSD with vmctl(8) present.
    """
    if __grains__["os"] == "OpenBSD" and salt.utils.path.which("vmctl"):
        return True

    return (
        False,
        "The vmm execution module cannot be loaded: either the system is not OpenBSD or"
        " the vmctl binary was not found",
    )


def _id_to_name(id):
    """
    Lookup the name associated with a VM id.
    """
    vm = status(id=id)
    if vm == {}:
        return None
    else:
        return vm["name"]


def create_disk(name, size):
    """
    Create a VMM disk with the specified `name` and `size`.

    size:
        Size in megabytes, or use a specifier such as M, G, T.

    CLI Example:

    .. code-block:: bash

        salt '*' vmctl.create_disk /path/to/disk.img size=10G
    """
    ret = False
    cmd = f"vmctl create {name} -s {size}"

    result = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)

    if result["retcode"] == 0:
        ret = True
    else:
        raise CommandExecutionError(
            "Problem encountered creating disk image",
            info={"errors": [result["stderr"]], "changes": ret},
        )

    return ret


def load(path):
    """
    Load additional configuration from the specified file.

    path
        Path to the configuration file.

    CLI Example:

    .. code-block:: bash

        salt '*' vmctl.load path=/etc/vm.switches.conf
    """
    ret = False
    cmd = f"vmctl load {path}"
    result = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)
    if result["retcode"] == 0:
        ret = True
    else:
        raise CommandExecutionError(
            "Problem encountered running vmctl",
            info={"errors": [result["stderr"]], "changes": ret},
        )

    return ret


def reload():
    """
    Remove all stopped VMs and reload configuration from the default configuration file.

    CLI Example:

    .. code-block:: bash

        salt '*' vmctl.reload
    """
    ret = False
    cmd = "vmctl reload"
    result = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)
    if result["retcode"] == 0:
        ret = True
    else:
        raise CommandExecutionError(
            "Problem encountered running vmctl",
            info={"errors": [result["stderr"]], "changes": ret},
        )

    return ret


def reset(all=False, vms=False, switches=False):
    """
    Reset the running state of VMM or a subsystem.

    all:
        Reset the running state.

    switches:
        Reset the configured switches.

    vms:
        Reset and terminate all VMs.

    CLI Example:

    .. code-block:: bash

        salt '*' vmctl.reset all=True
    """
    ret = False
    cmd = ["vmctl", "reset"]

    if all:
        cmd.append("all")
    elif vms:
        cmd.append("vms")
    elif switches:
        cmd.append("switches")

    result = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)
    if result["retcode"] == 0:
        ret = True
    else:
        raise CommandExecutionError(
            "Problem encountered running vmctl",
            info={"errors": [result["stderr"]], "changes": ret},
        )

    return ret


def start(
    name=None,
    id=None,
    bootpath=None,
    disk=None,
    disks=None,
    local_iface=False,
    memory=None,
    nics=0,
    switch=None,
):
    """
    Starts a VM defined by the specified parameters.
    When both a name and id are provided, the id is ignored.

    name:
        Name of the defined VM.

    id:
        VM id.

    bootpath:
        Path to a kernel or BIOS image to load.

    disk:
        Path to a single disk to use.

    disks:
        List of multiple disks to use.

    local_iface:
        Whether to add a local network interface. See "LOCAL INTERFACES"
        in the vmctl(8) manual page for more information.

    memory:
        Memory size of the VM specified in megabytes.

    switch:
        Add a network interface that is attached to the specified
        virtual switch on the host.

    CLI Example:

    .. code-block:: bash

        salt '*' vmctl.start 2   # start VM with id 2
        salt '*' vmctl.start name=web1 bootpath='/bsd.rd' nics=2 memory=512M disk='/disk.img'
    """
    ret = {"changes": False, "console": None}
    cmd = ["vmctl", "start"]

    if not (name or id):
        raise SaltInvocationError('Must provide either "name" or "id"')
    elif name:
        cmd.append(name)
    else:
        cmd.append(id)
        name = _id_to_name(id)

    if nics > 0:
        cmd.append(f"-i {nics}")

    # Paths cannot be appended as otherwise the inserted whitespace is treated by
    # vmctl as being part of the path.
    if bootpath:
        cmd.extend(["-b", bootpath])

    if memory:
        cmd.append(f"-m {memory}")

    if switch:
        cmd.append(f"-n {switch}")

    if local_iface:
        cmd.append("-L")

    if disk and disks:
        raise SaltInvocationError('Must provide either "disks" or "disk"')

    if disk:
        cmd.extend(["-d", disk])

    if disks:
        cmd.extend(["-d", x] for x in disks)

    # Before attempting to define a new VM, make sure it doesn't already exist.
    # Otherwise return to indicate nothing was changed.
    if len(cmd) > 3:
        vmstate = status(name)
        if vmstate:
            ret["comment"] = "VM already exists and cannot be redefined"
            return ret

    result = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)

    if result["retcode"] == 0:
        ret["changes"] = True
        m = re.match(r".*successfully, tty (\/dev.*)", result["stderr"])
        if m:
            ret["console"] = m.groups()[0]
        else:
            m = re.match(r".*Operation already in progress$", result["stderr"])
            if m:
                ret["changes"] = False
    else:
        raise CommandExecutionError(
            "Problem encountered running vmctl",
            info={"errors": [result["stderr"]], "changes": ret},
        )

    return ret


def status(name=None, id=None):
    """
    List VMs running on the host, or only the VM specified by ``id``.  When
    both a name and id are provided, the id is ignored.

    name:
        Name of the defined VM.

    id:
        VM id.

    CLI Example:

    .. code-block:: bash

        salt '*' vmctl.status           # to list all VMs
        salt '*' vmctl.status name=web1 # to get a single VM
    """
    ret = {}
    cmd = ["vmctl", "status"]

    result = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)

    if result["retcode"] != 0:
        raise CommandExecutionError(
            "Problem encountered running vmctl",
            info={"error": [result["stderr"]], "changes": ret},
        )

    # Grab the header and save it with the lowercase names.
    header = result["stdout"].splitlines()[0].split()
    header = [x.lower() for x in header]

    # A VM can be in one of the following states (from vmm.c:vcpu_state_decode())
    # - stopped
    # - running
    # - requesting termination
    # - terminated
    # - unknown

    for line in result["stdout"].splitlines()[1:]:
        data = line.split()
        vm = dict(list(zip(header, data)))
        vmname = vm.pop("name")
        if vm["pid"] == "-":
            # If the VM has no PID it's not running.
            vm["state"] = "stopped"
        elif vmname and data[-2] == "-":
            # When a VM does have a PID and the second to last field is a '-', it's
            # transitioning to another state. A VM name itself cannot contain a
            # '-' so it's safe to split on '-'.
            vm["state"] = data[-1]
        else:
            vm["state"] = "running"

        # When the status is requested of a single VM (by name) which is stopping,
        # vmctl doesn't print the status line. So we'll parse the full list and
        # return when we've found the requested VM.
        if id and int(vm["id"]) == id:
            return {vmname: vm}
        elif name and vmname == name:
            return {vmname: vm}
        else:
            ret[vmname] = vm

    # Assert we've not come this far when an id or name have been provided. That
    # means the requested VM does not exist.
    if id or name:
        return {}

    return ret


def stop(name=None, id=None):
    """
    Stop (terminate) the VM identified by the given id or name.
    When both a name and id are provided, the id is ignored.

    name:
        Name of the defined VM.

    id:
        VM id.

    CLI Example:

    .. code-block:: bash

        salt '*' vmctl.stop name=alpine
    """
    ret = {}
    cmd = ["vmctl", "stop"]

    if not (name or id):
        raise SaltInvocationError('Must provide either "name" or "id"')
    elif name:
        cmd.append(name)
    else:
        cmd.append(id)

    result = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)

    if result["retcode"] == 0:
        if re.match("^vmctl: sent request to terminate vm.*", result["stderr"]):
            ret["changes"] = True
        else:
            ret["changes"] = False
    else:
        raise CommandExecutionError(
            "Problem encountered running vmctl",
            info={"errors": [result["stderr"]], "changes": ret},
        )

    return ret



"""
Tests for file.touch function
"""

import pytest

from salt.exceptions import SaltInvocationError

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.fixture(scope="module")
def file(modules):
    return modules.file


def test_touch(file, tmp_path):
    """
    Test touch with defaults
    """
    target = tmp_path / "test.file"
    file.touch(str(target))
    assert target.exists()


def test_touch_error_atime(file, tmp_path):
    """
    Test touch with non int input
    """
    target = tmp_path / "test.file"
    with pytest.raises(SaltInvocationError) as exc:
        file.touch(str(target), atime="string")
    assert "atime and mtime must be integers" in exc.value.message


def test_touch_error_mtime(file, tmp_path):
    """
    Test touch with non int input
    """
    target = tmp_path / "test.file"
    with pytest.raises(SaltInvocationError) as exc:
        file.touch(str(target), mtime="string")
    assert "atime and mtime must be integers" in exc.value.message


def test_touch_atime(file, tmp_path):
    """
    Test touch with defaults
    """
    target = tmp_path / "test.file"
    file.touch(str(target), atime=123)
    assert target.stat().st_atime == 123


def test_touch_atime_zero(file, tmp_path):
    """
    Test touch with defaults
    """
    target = tmp_path / "test.file"
    file.touch(str(target), atime=0)
    assert target.stat().st_atime == 0


def test_touch_mtime(file, tmp_path):
    """
    Test touch with defaults
    """
    target = tmp_path / "test.file"
    file.touch(str(target), mtime=234)
    assert target.stat().st_mtime == 234


def test_touch_mtime_zero(file, tmp_path):
    """
    Test touch with defaults
    """
    target = tmp_path / "test.file"
    file.touch(str(target), mtime=0)
    assert target.stat().st_mtime == 0


def test_touch_atime_mtime(file, tmp_path):
    """
    Test touch with defaults
    """
    target = tmp_path / "test.file"
    file.touch(str(target), atime=456, mtime=789)
    assert target.stat().st_atime == 456
    assert target.stat().st_mtime == 789



import pytest
from salt.utils.templates import render_mako_tmpl

pytest.importorskip("mako")


def test_render_mako_sanity(render_context):
    tmpl = """OK"""
    res = render_mako_tmpl(tmpl, render_context)
    assert res == "OK"


def test_render_mako_evaluate(render_context):
    tmpl = """${ "OK" }"""
    res = render_mako_tmpl(tmpl, render_context)
    assert res == "OK"


def test_render_mako_evaluate_multi(render_context):
    tmpl = """
    % if 1:
    OK
    % endif
    """
    res = render_mako_tmpl(tmpl, render_context)
    stripped = res.strip()
    assert stripped == "OK"


def test_render_mako_variable(render_context):
    tmpl = """${ var }"""
    render_context["var"] = "OK"
    res = render_mako_tmpl(tmpl, render_context)
    assert res == "OK"



"""
Management of OpenStack Keystone Endpoints
==========================================

.. versionadded:: 2018.3.0

:depends: shade
:configuration: see :py:mod:`salt.modules.keystoneng` for setup instructions

Example States

.. code-block:: yaml

    create endpoint:
      keystone_endpoint.present:
        - name: public
        - url: https://example.org:9292
        - region: RegionOne
        - service_name: glance

    destroy endpoint:
      keystone_endpoint.absent:
        - name: public
        - url: https://example.org:9292
        - region: RegionOne
        - service_name: glance

    create multiple endpoints:
      keystone_endpoint.absent:
        - names:
            - public
            - admin
            - internal
        - url: https://example.org:9292
        - region: RegionOne
        - service_name: glance
"""

__virtualname__ = "keystone_endpoint"


def __virtual__():
    if "keystoneng.endpoint_get" in __salt__:
        return __virtualname__
    return (
        False,
        "The keystoneng execution module failed to load: shade python module is not"
        " available",
    )


def _common(ret, name, service_name, kwargs):
    """
    Returns: tuple whose first element is a bool indicating success or failure
             and the second element is either a ret dict for salt or an object
    """
    if "interface" not in kwargs and "public_url" not in kwargs:
        kwargs["interface"] = name
    service = __salt__["keystoneng.service_get"](name_or_id=service_name)

    if not service:
        ret["comment"] = "Cannot find service"
        ret["result"] = False
        return (False, ret)

    filters = kwargs.copy()
    filters.pop("enabled", None)
    filters.pop("url", None)
    filters["service_id"] = service.id
    kwargs["service_name_or_id"] = service.id
    endpoints = __salt__["keystoneng.endpoint_search"](filters=filters)

    if len(endpoints) > 1:
        ret["comment"] = "Multiple endpoints match criteria"
        ret["result"] = False
        return ret
    endpoint = endpoints[0] if endpoints else None
    return (True, endpoint)


def present(name, service_name, auth=None, **kwargs):
    """
    Ensure an endpoint exists and is up-to-date

    name
        Interface name

    url
        URL of the endpoint

    service_name
        Service name or ID

    region
        The region name to assign the endpoint

    enabled
        Boolean to control if endpoint is enabled
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    kwargs = __utils__["args.clean_kwargs"](**kwargs)

    __salt__["keystoneng.setup_clouds"](auth)

    success, val = _, endpoint = _common(ret, name, service_name, kwargs)
    if not success:
        return val

    if not endpoint:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = kwargs
            ret["comment"] = "Endpoint will be created."
            return ret

        # NOTE(SamYaple): Endpoints are returned as a list which can contain
        # several items depending on the options passed
        endpoints = __salt__["keystoneng.endpoint_create"](**kwargs)
        if len(endpoints) == 1:
            ret["changes"] = endpoints[0]
        else:
            for i, endpoint in enumerate(endpoints):
                ret["changes"][i] = endpoint
        ret["comment"] = "Created endpoint"
        return ret

    changes = __salt__["keystoneng.compare_changes"](endpoint, **kwargs)
    if changes:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = changes
            ret["comment"] = "Endpoint will be updated."
            return ret

        kwargs["endpoint_id"] = endpoint.id
        __salt__["keystoneng.endpoint_update"](**kwargs)
        ret["changes"].update(changes)
        ret["comment"] = "Updated endpoint"

    return ret


def absent(name, service_name, auth=None, **kwargs):
    """
    Ensure an endpoint does not exists

    name
        Interface name

    url
        URL of the endpoint

    service_name
        Service name or ID

    region
        The region name to assign the endpoint
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    __salt__["keystoneng.setup_clouds"](auth)

    success, val = _, endpoint = _common(ret, name, service_name, kwargs)
    if not success:
        return val

    if endpoint:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = {"id": endpoint.id}
            ret["comment"] = "Endpoint will be deleted."
            return ret

        __salt__["keystoneng.endpoint_delete"](id=endpoint.id)
        ret["changes"]["id"] = endpoint.id
        ret["comment"] = "Deleted endpoint"

    return ret



"""
Decorators for salt.state

:codeauthor: :email:`Bo Maryniuk (bo@suse.de)`
"""

import logging

import salt.utils.stringutils
from salt.exceptions import SaltException

log = logging.getLogger(__name__)


class OutputUnifier:
    def __init__(self, *policies):
        self.policies = []
        for pls in policies:
            if not hasattr(self, pls):
                raise SaltException(f"Unknown policy: {pls}")
            else:
                self.policies.append(getattr(self, pls))

    def _run_policies(self, data):
        for pls in self.policies:
            try:
                data = pls(data)
            except Exception as exc:  # pylint: disable=broad-except
                log.debug(
                    "An exception occurred in this state: %s",
                    exc,
                    exc_info_on_loglevel=logging.DEBUG,
                )
                data = {
                    "result": False,
                    "name": "later",
                    "changes": {},
                    "comment": f"An exception occurred in this state: {exc}",
                }
        return data

    def __call__(self, func):
        def _func(*args, **kwargs):
            result = func(*args, **kwargs)
            sub_state_run = None
            if isinstance(result, dict):
                sub_state_run = result.get("sub_state_run", ())
            result = self._run_policies(result)
            if sub_state_run:
                result["sub_state_run"] = [
                    self._run_policies(sub_state_data)
                    for sub_state_data in sub_state_run
                ]
            return result

        return _func

    def content_check(self, result):
        """
        Checks for specific types in the state output.
        Raises an Exception in case particular rule is broken.

        :param result:
        :return:
        """
        if not isinstance(result, dict):
            err_msg = "Malformed state return. Data must be a dictionary type."
        elif not isinstance(result.get("changes"), dict):
            err_msg = "'Changes' should be a dictionary."
        else:
            missing = []
            for val in ["name", "result", "changes", "comment"]:
                if val not in result:
                    missing.append(val)
            if missing:
                err_msg = "The following keys were not present in the state return: {}.".format(
                    ", ".join(missing)
                )
            else:
                err_msg = None

        if err_msg:
            raise SaltException(err_msg)

        for sub_state in result.get("sub_state_run", ()):
            self.content_check(sub_state)

        return result

    def unify(self, result):
        """
        While comments as a list are allowed,
        comments needs to be strings for backward compatibility.
        See such claim here: https://github.com/saltstack/salt/pull/43070

        Rules applied:
          - 'comment' is joined into a multi-line string, in case the value is a list.
          - 'result' should be always either True, False or None.

        :param result:
        :return:
        """
        if isinstance(result.get("comment"), list):
            result["comment"] = "\n".join(
                [salt.utils.stringutils.to_unicode(elm) for elm in result["comment"]]
            )
        if result.get("result") is not None:
            result["result"] = bool(result["result"])

        return result



"""
Configuration disposable regularly scheduled tasks for at.
==========================================================

The at state can be add disposable regularly scheduled tasks for your system.
"""

import logging

log = logging.getLogger(__name__)


def __virtual__():
    """
    Most everything has the ability to support at(1)
    """
    if "at.at" in __salt__:
        return True
    return (False, "at module could not be loaded")


def present(name, timespec, tag=None, user=None, job=None, unique_tag=False):
    """
    .. versionchanged:: 2017.7.0

    Add a job to queue.

    job : string
        Command to run.

    timespec : string
        The 'timespec' follows the format documented in the at(1) manpage.

    tag : string
        Make a tag for the job.

    user : string
        The user to run the at job
        .. versionadded:: 2014.1.4

    unique_tag : boolean
        If set to True job will not be added if a job with the tag exists.
        .. versionadded:: 2017.7.0

    .. code-block:: yaml

        rose:
          at.present:
            - job: 'echo "I love saltstack" > love'
            - timespec: '9:09 11/09/13'
            - tag: love
            - user: jam

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # if job is missing, use name
    if not job:
        job = name

    # quick return on test=True
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "job {} added and will run on {}".format(
            job,
            timespec,
        )
        return ret

    # quick return if unique_tag and job exists
    if unique_tag:
        if not tag:
            ret["result"] = False
            ret["comment"] = "no tag provided and unique_tag is set to True"
            return ret
        elif len(__salt__["at.jobcheck"](tag=tag)["jobs"]) > 0:
            ret["comment"] = f"atleast one job with tag {tag} exists."
            return ret

    # create job
    if user:
        luser = __salt__["user.info"](user)
        if not luser:
            ret["result"] = False
            ret["comment"] = f"user {user} does not exists"
            return ret
        ret["comment"] = "job {} added and will run as {} on {}".format(
            job,
            user,
            timespec,
        )
        res = __salt__["at.at"](
            timespec,
            job,
            tag=tag,
            runas=user,
        )
    else:
        ret["comment"] = "job {} added and will run on {}".format(
            job,
            timespec,
        )
        res = __salt__["at.at"](
            timespec,
            job,
            tag=tag,
        )

    # set ret['changes']
    if "jobs" in res and len(res["jobs"]) > 0:
        ret["changes"] = res["jobs"][0]
    if "error" in res:
        ret["result"] = False
        ret["comment"] = res["error"]

    return ret


def absent(name, jobid=None, **kwargs):
    """
    .. versionchanged:: 2017.7.0

    Remove a job from queue

    jobid: string|int
        Specific jobid to remove

    tag : string
        Job's tag

    runas : string
        Runs user-specified jobs

    kwargs
        Addition kwargs can be provided to filter jobs.
        See output of `at.jobcheck` for more.

    .. code-block:: yaml

        example1:
          at.absent:

    .. warning::
        this will remove all jobs!

    .. code-block:: yaml

        example2:
          at.absent:
            - year: 13

    .. code-block:: yaml

        example3:
          at.absent:
            - tag: rose

    .. code-block:: yaml

        example4:
          at.absent:
            - tag: rose
            - day: 13
            - hour: 16

    .. code-block:: yaml

        example5:
          at.absent:
            - jobid: 4

    .. note:
        all other filters are ignored and only job with id 4 is removed
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # limit was never support
    if "limit" in kwargs:
        ret["comment"] = f"limit parameter not supported {name}"
        ret["result"] = False
        return ret

    # quick return on test=True
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "removed ? job(s)"
        return ret

    # remove specific job
    if jobid:
        jobs = __salt__["at.atq"](jobid)
        if "jobs" in jobs and len(jobs["jobs"]) == 0:
            ret["result"] = True
            ret["comment"] = f"job with id {jobid} not present"
            return ret
        elif "jobs" in jobs and len(jobs["jobs"]) == 1:
            if "job" in jobs["jobs"][0] and jobs["jobs"][0]["job"]:
                res = __salt__["at.atrm"](jobid)
                ret["result"] = jobid in res["jobs"]["removed"]
                if ret["result"]:
                    ret["comment"] = "job with id {jobid} was removed".format(
                        jobid=jobid
                    )
                else:
                    ret["comment"] = "failed to remove job with id {jobid}".format(
                        jobid=jobid
                    )
                ret["changes"]["removed"] = res["jobs"]["removed"]
                return ret
        else:
            ret["result"] = False
            ret["comment"] = (
                "more than one job was return for job with id {jobid}".format(
                    jobid=jobid
                )
            )
            return ret

    # remove jobs based on filter
    if kwargs:
        # we pass kwargs to at.jobcheck
        opts = list(
            list(
                map(str, [j["job"] for j in __salt__["at.jobcheck"](**kwargs)["jobs"]])
            )
        )
        res = __salt__["at.atrm"](*opts)
    else:
        # arguments to filter with, removing everything!
        res = __salt__["at.atrm"]("all")

    if len(res["jobs"]["removed"]) > 0:
        ret["changes"]["removed"] = res["jobs"]["removed"]
    ret["comment"] = "removed {count} job(s)".format(count=len(res["jobs"]["removed"]))
    return ret


def watch(name, timespec, tag=None, user=None, job=None, unique_tag=False):
    """
    .. versionadded:: 2017.7.0

    Add an at job if trigger by watch

    job : string
        Command to run.

    timespec : string
        The 'timespec' follows the format documented in the at(1) manpage.

    tag : string
        Make a tag for the job.

    user : string
        The user to run the at job
        .. versionadded:: 2014.1.4

    unique_tag : boolean
        If set to True job will not be added if a job with the tag exists.
        .. versionadded:: 2017.7.0

    .. code-block:: yaml

        minion_restart:
          at.watch:
            - job: 'salt-call --local service.restart salt-minion'
            - timespec: 'now +1 min'
            - tag: minion_restart
            - unique_tag: trye
            - watch:
                - file: /etc/salt/minion

    """
    return {"name": name, "changes": {}, "result": True, "comment": ""}


def mod_watch(name, **kwargs):
    """
    The at watcher, called to invoke the watch command.

    .. note::
        This state exists to support special handling of the ``watch``
        :ref:`requisite <requisites>`. It should not be called directly.

        Parameters for this function should be set by the state being triggered.

    name
        The name of the atjob

    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if kwargs["sfun"] == "watch":
        for p in ["sfun", "__reqs__"]:
            del kwargs[p]
        kwargs["name"] = name
        ret = present(**kwargs)

    return ret
